import pkgutil
import sys
import os
import types
import warnings
import textwrap
from functools import partial
from operator import attrgetter
from math import log10

from typing import Optional, Union, List, cast

from AnyQt.QtWidgets import (
    QWidget, QDialog, QVBoxLayout, QSizePolicy, QApplication, QStyle,
    QSplitter, QSplitterHandle, QPushButton, QStatusBar,
    QProgressBar, QAction, QFrame, QStyleOption, QHBoxLayout, QMenuBar, QMenu,
    QWIDGETSIZE_MAX
)
from AnyQt.QtCore import (
    Qt, QObject, QEvent, QRect, QMargins, QByteArray, QDataStream, QBuffer,
    QSettings, QUrl, QThread, QTimer, QSize, QPoint, QLine,
    pyqtSignal as Signal
)
from AnyQt.QtGui import (
    QIcon, QKeySequence, QDesktopServices, QPainter, QColor, QPen, QKeyEvent,
    QActionEvent
)

from orangecanvas.gui.svgiconengine import StyledSvgIconEngine

from orangewidget import settings, gui
from orangewidget.report import Report
from orangewidget.gui import OWComponent, VerticalScrollArea
from orangewidget.io import ClipboardFormat, ImgFormat
from orangewidget.settings import SettingsHandler
from orangewidget.utils import saveplot, getdeepattr
from orangewidget.utils.messagewidget import InOutStateWidget
from orangewidget.utils.progressbar import ProgressBarMixin
from orangewidget.utils.messages import (
    WidgetMessagesMixin, UnboundMsg, MessagesWidget
)
from orangewidget.utils.signals import (
    WidgetSignalsMixin, Input, Output, MultiInput,
    InputSignal, OutputSignal,
    Default, NonDefault, Single, Multiple, Dynamic, Explicit
)  # pylint: disable=unused-import
from orangewidget.utils.overlay import MessageOverlayWidget, OverlayWidget
from orangewidget.utils.buttons import SimpleButton
from orangewidget.utils import dropdown_popup_geometry

# Msg is imported and renamed, so widgets can import it from this module rather
# than the one with the mixin (orangewidget.utils.messages).
Msg = UnboundMsg


__all__ = [
    "OWBaseWidget", "Input", "Output", "MultiInput",
    "Message", "Msg", "StateInfo",
]


def _load_styled_icon(name):
    return QIcon(StyledSvgIconEngine(pkgutil.get_data(__name__, "icons/" + name)))


class Message:
    """
    A user message.

    :param str text: Message text
    :param str persistent_id:
        A persistent message id.
    :param icon: Message icon
    :type icon: QIcon or QStyle.StandardPixmap
    :param str moreurl:
        An url to open when a user clicks a 'Learn more' button.

    .. seealso:: :const:`OWBaseWidget.UserAdviceMessages`
    """
    #: QStyle.SP_MessageBox* pixmap enums repeated for easier access
    Question = QStyle.SP_MessageBoxQuestion
    Information = QStyle.SP_MessageBoxInformation
    Warning = QStyle.SP_MessageBoxWarning
    Critical = QStyle.SP_MessageBoxCritical

    def __init__(self, text, persistent_id, icon=None, moreurl=None):
        assert isinstance(text, str)
        assert isinstance(icon, (type(None), QIcon, QStyle.StandardPixmap))
        assert persistent_id is not None
        self.text = text
        self.icon = icon
        self.moreurl = moreurl
        self.persistent_id = persistent_id


def _asmappingproxy(mapping):
    if isinstance(mapping, types.MappingProxyType):
        return mapping
    else:
        return types.MappingProxyType(mapping)


class WidgetMetaClass(type(QDialog)):
    """Meta class for widgets. If the class definition does not have a
       specific settings handler, the meta class provides a default one
       that does not handle contexts. Then it scans for any attributes
       of class settings.Setting: the setting is stored in the handler and
       the value of the attribute is replaced with the default."""

    #noinspection PyMethodParameters
    # pylint: disable=bad-classmethod-argument
    def __new__(mcs, name, bases, namespace, openclass=False, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)
        cls.convert_signals()
        if isinstance(cls.keywords, str):
            if "," in cls.keywords:
                cls.keywords = [kw.strip() for kw in cls.keywords.split(",")]
            else:
                cls.keywords = cls.keywords.split()
        if not cls.name: # not a widget
            return cls
        cls.settingsHandler = \
            SettingsHandler.create(cls, template=cls.settingsHandler)
        return cls

    @classmethod
    # pylint: disable=bad-classmethod-argument
    def __prepare__(mcs, name, bases, metaclass=None, openclass=False,
                    **kwargs):
        namespace = super().__prepare__(mcs, name, bases, metaclass, **kwargs)
        if not openclass:
            namespace["_final_class"] = True
        return namespace


# pylint: disable=too-many-instance-attributes
class OWBaseWidget(QDialog, OWComponent, Report, ProgressBarMixin,
                   WidgetMessagesMixin, WidgetSignalsMixin,
                   metaclass=WidgetMetaClass, openclass=True):
    """
    Base widget class in an Orange widget workflow.
    """

    #: Widget name, as presented in the Canvas.
    name: str = None

    #: Short widget description, displayed in canvas help tooltips.
    description: str = ""

    #: Widget icon path, relative to the defining module.
    icon: str = "icons/Unknown.png"

    class Inputs:
        """
        Define inputs in this nested class as class variables.
        (type `orangewidget.widget.Input`)

        Example::

            class Inputs:
                data = Input("Data", Table)

        Then, register input handler methods with decorators.

        Example::

            @Inputs.data
            def set_data(self, data):
                self.my_data = data
        """

    class Outputs:
        """
        Define outputs in this nested class as class variables.
        (type `orangewidget.widget.Output`)

        Example::

            class Outputs:
                data = Output("Data", Table)

        Then, send results to the output with its `send` method.

        Example::

            def commit(self):
                self.Outputs.data.send(self.my_data)
        """

    # -------------------------------------------------------------------------
    # Widget GUI Layout Settings
    # -------------------------------------------------------------------------

    #: Should the widget have basic layout?
    #: (if not, the rest of the GUI layout settings are ignored)
    want_basic_layout = True

    #: Should the widget construct a `controlArea`?
    want_control_area = True

    #: Should the widget construct a `mainArea`?
    #: (a resizable area to the right of the `controlArea`)
    want_main_area = True

    #: Should the widget construct a `message_bar`?
    #: (if not, make sure you show warnings/errors in some other way)
    want_message_bar = True

    #: Should the widget's window be resizeable?
    #: (if not, the widget will derive a fixed size constraint from its layout)
    resizing_enabled = True

    #: Should the widget remember its window position/size?
    save_position = True

    #: Orientation of the buttonsArea box; valid only if
    #: `want_control_area` is `True`. Possible values are Qt.Horizontal,
    #: Qt.Vertical and None for no buttons area
    buttons_area_orientation = Qt.Horizontal

    #: A list of advice messages to display to the user.
    #: (when a widget is first shown a message from this list is selected
    #: for display. If a user accepts (clicks 'Ok. Got it') the choice is
    #: recorded and the message is never shown again (closing the message
    #: will not mark it as seen). Messages can be displayed again by pressing
    #: Shift + F1)
    UserAdviceMessages: List[Message] = []

    # -------------------------------------------------------------------------
    # Miscellaneous Options
    # -------------------------------------------------------------------------

    #: Version of the settings representation
    #: (subclasses should increase this number when they make breaking
    #: changes to settings representation (a settings that used to store
    #: int now stores string) and handle migrations in migrate and
    #: migrate_context settings)
    settings_version: int = 1

    #: Signal emitted before settings are packed and saved.
    #: (gives you a chance to sync state to Setting values)
    settingsAboutToBePacked = Signal()

    #: Settings handler, can be overridden for context handling.
    settingsHandler: SettingsHandler = None

    #: Widget keywords, used for finding it in the quick menu.
    keywords: Union[str, List[str]] = []

    #: Widget priority, used for sorting within a category.
    priority: int = sys.maxsize

    #: Short name for widget, displayed in toolbox.
    #: (set this if the widget's conventional name is long)
    short_name: str = None

    #: A list of widget IDs that this widget replaces in older workflows.
    replaces: List[str] = None

    #: Widget painted by `Save graph` button
    graph_name: str = None
    graph_writers: List[ImgFormat] = [f for f in ImgFormat.formats
                                  if getattr(f, 'write_image', None)
                                  and getattr(f, "EXTENSIONS", None)]

    #: Explicitly set widget category,
    #: should it not already be part of a package.
    category: str = None

    #: Ratio between width and height for mainArea widgets,
    #: set to None for super().sizeHint()
    mainArea_width_height_ratio: Optional[float] = 1.1

    # -------------------------------------------------------------------------
    # Private Interface
    # -------------------------------------------------------------------------

    # Custom widget id, kept for backward compatibility
    id = None

    # A list of published input definitions.
    # (conventionally generated from Inputs nested class)
    inputs = []

    # A list of published output definitions.
    # (conventionally generated from Outputs nested class)
    outputs = []

    contextAboutToBeOpened = Signal(object)
    contextOpened = Signal()
    contextClosed = Signal()
    openVisualSettingsClicked = Signal()

    # Signals have to be class attributes and cannot be inherited,
    # say from a mixin. This has something to do with the way PyQt binds them
    progressBarValueChanged = Signal(float)
    messageActivated = Signal(Msg)
    messageDeactivated = Signal(Msg)

    savedWidgetGeometry = settings.Setting(None)
    controlAreaVisible = settings.Setting(True, schema_only=True)

    __report_action = None  # type: Optional[QAction]
    __save_image_action = None  # type: Optional[QAction]
    __reset_action = None  # type: Optional[QAction]
    __visual_settings_action = None  # type: Optional[QAction]
    __menuBar = None  # type: QMenuBar

    # pylint: disable=protected-access, access-member-before-definition
    def __new__(cls, *args, captionTitle=None, **kwargs):
        self = super().__new__(cls, None, cls.get_flags())
        QDialog.__init__(self, None, self.get_flags())
        OWComponent.__init__(self)
        WidgetMessagesMixin.__init__(self)
        WidgetSignalsMixin.__init__(self)

        # handle deprecated left_side_scrolling
        if hasattr(self, 'left_side_scrolling'):
            warnings.warn(
                "'OWBaseWidget.left_side_scrolling' is deprecated.",
                DeprecationWarning
            )

        stored_settings = kwargs.get('stored_settings', None)
        if self.settingsHandler:
            self.settingsHandler.initialize(self, stored_settings)

        self.signalManager = kwargs.get('signal_manager', None)
        self.__env = _asmappingproxy(kwargs.get("env", {}))

        self.graphButton = None
        self.report_button = None

        captionTitle = self.name if captionTitle is None else captionTitle

        # must be set without invoking setCaption
        self.captionTitle = captionTitle
        self.setWindowTitle(captionTitle)

        self.setFocusPolicy(Qt.StrongFocus)

        # flag indicating if the widget's position was already restored
        self.__was_restored = False
        # flag indicating the widget is still expecting the first show event.
        self.__was_shown = False

        self.__statusMessage = ""
        self.__info_ns = None  # type: Optional[StateInfo]
        self.__msgwidget = None  # type: Optional[MessageOverlayWidget]
        self.__msgchoice = 0
        self.__statusbar = None  # type: Optional[QStatusBar]
        self.__statusbar_action = None  # type: Optional[QAction]
        self.__menubar_action = None
        self.__menubar_visible_timer = None

        # this action is enabled by the canvas framework
        self.__help_action = QAction(
            "Help", self, objectName="action-help", toolTip="Show help",
            enabled=False, shortcut=QKeySequence(Qt.Key_F1)
        )
        self.__report_action = QAction(
            "Report", self, objectName="action-report",
            toolTip="Create and display a report",
            enabled=False, visible=False,
            shortcut=QKeySequence("alt+r")
        )
        if hasattr(self, "send_report"):
            self.__report_action.triggered.connect(self.show_report)
            self.__report_action.setEnabled(True)
            self.__report_action.setVisible(True)

        self.__save_image_action = QAction(
            "Save Image", self, objectName="action-save-image",
            toolTip="Save image",
            shortcut=QKeySequence("alt+s"),
        )
        self.__save_image_action.triggered.connect(self.save_graph)
        self.__save_image_action.setEnabled(bool(self.graph_name))
        self.__save_image_action.setVisible(bool(self.graph_name))

        self.__reset_action = QAction(
            "Reset", self, objectName="action-reset-settings",
            toolTip="Reset settings to default state",
            enabled=False, visible=False,
        )
        if hasattr(self, "reset_settings"):
            self.__reset_action.triggered.connect(self.reset_settings)
            self.__reset_action.setEnabled(True)
            self.__reset_action.setVisible(True)

        self.__visual_settings_action = QAction(
            "Show View Options", self, objectName="action-visual-settings",
            toolTip="Show View Options",
            enabled=False, visible=False,
        )
        self.__visual_settings_action.triggered.connect(
            self.openVisualSettingsClicked)
        if hasattr(self, "set_visual_settings"):
            self.__visual_settings_action.setEnabled(True)
            self.__visual_settings_action.setVisible(True)

        self.addAction(self.__help_action)

        self.__copy_action = QAction(
            "Copy to Clipboard", self, objectName="action-copy-to-clipboard",
            shortcut=QKeySequence.Copy, enabled=False, visible=False
        )
        self.__copy_action.triggered.connect(self.copy_to_clipboard)
        if type(self).copy_to_clipboard != OWBaseWidget.copy_to_clipboard \
                or self.graph_name is not None:
            self.__copy_action.setEnabled(True)
            self.__copy_action.setVisible(True)
            self.__copy_action.setText("Copy Image to Clipboard")

        # macOS Minimize action
        self.__minimize_action = QAction(
            "Minimize", self, shortcut=QKeySequence("ctrl+m")
        )
        self.__minimize_action.triggered.connect(self.showMinimized)
        # macOS Close window action
        self.__close_action = QAction(
            "Close", self, objectName="action-close-window",
            shortcut=QKeySequence("ctrl+w")
        )
        self.__close_action.triggered.connect(self.hide)

        settings = QSettings()
        settings.beginGroup(__name__ + "/menubar")
        self.__menubar = mb = QMenuBar(self)
        # do we have a native menubar
        nativemb = mb.isNativeMenuBar()
        if nativemb:
            # force non native mb via. settings override
            nativemb = settings.value(
                "use-native", defaultValue=nativemb, type=bool
            )
        mb.setNativeMenuBar(nativemb)
        if not nativemb:
            # without native menu bar configure visibility
            mbvisible = settings.value(
                "visible", defaultValue=False, type=bool
            )
            mb.setVisible(mbvisible)
            self.__menubar_action = QAction(
                "Show Menu Bar",  self, objectName="action-show-menu-bar",
                checkable=True,
                shortcut=QKeySequence("ctrl+shift+m")
            )
            self.__menubar_action.setChecked(mbvisible)
            self.__menubar_action.triggered[bool].connect(
                self.__setMenuBarVisible
            )
            self.__menubar_visible_timer = QTimer(
                self, objectName="menu-bar-visible-timer", singleShot=True,
                interval=settings.value(
                    "alt-key-timeout", defaultValue=50, type=int,
                )
            )
            self.__menubar_visible_timer.timeout.connect(
                self.__menuBarVisibleTimeout
            )
            self.addAction(self.__menubar_action)

        fileaction = mb.addMenu(_Menu("&File", mb, objectName="menu-file"))
        fileaction.setVisible(False)
        fileaction.menu().addSeparator()
        fileaction.menu().addAction(self.__report_action)
        fileaction.menu().addAction(self.__save_image_action)
        fileaction.menu().addAction(self.__reset_action)
        editaction = mb.addMenu(_Menu("&Edit", mb, objectName="menu-edit"))
        editaction.setVisible(False)

        editaction.menu().addAction(self.__copy_action)
        if sys.platform == "darwin" and mb.isNativeMenuBar():
            # QTBUG-17291
            editaction.menu().addAction(
                QAction(
                    "Cut", self, enabled=False,
                    shortcut=QKeySequence(QKeySequence.Cut),
            ))
            editaction.menu().addAction(
                QAction(
                    "Copy", self, enabled=False,
                    shortcut=QKeySequence(QKeySequence.Copy),
            ))
            editaction.menu().addAction(
                QAction(
                    "Paste", self, enabled=False,
                    shortcut=QKeySequence(QKeySequence.Paste),
            ))
            editaction.menu().addAction(
                QAction(
                    "Select All", self, enabled=False,
                    shortcut=QKeySequence(QKeySequence.SelectAll),
            ))

        viewaction = mb.addMenu(_Menu("&View", mb, objectName="menu-view"))
        viewaction.setVisible(False)
        viewaction.menu().addAction(self.__visual_settings_action)
        windowaction = mb.addMenu(_Menu("&Window", mb, objectName="menu-window"))
        windowaction.setVisible(False)

        if sys.platform == "darwin":
            windowaction.menu().addAction(self.__close_action)
            windowaction.menu().addAction(self.__minimize_action)
            windowaction.menu().addSeparator()

        helpaction = mb.addMenu(_Menu("&Help", mb, objectName="help-menu"))
        helpaction.menu().addAction(self.__help_action)

        self.left_side = None
        self.controlArea = self.mainArea = self.buttonsArea = None

        self.__splitter = None
        if self.want_basic_layout:
            self.set_basic_layout()
            self.update_summaries()
            self.layout().setMenuBar(mb)

        self.__quick_help_action = QAction(
            "Quick Help Tip", self, objectName="action-quick-help-tip",
            shortcut=QKeySequence("shift+f1")
        )
        self.__quick_help_action.setEnabled(bool(self.UserAdviceMessages))
        self.__quick_help_action.setVisible(bool(self.UserAdviceMessages))
        self.__quick_help_action.triggered.connect(self.__quicktip)
        helpaction.menu().addAction(self.__quick_help_action)

        if self.__splitter is not None and self.__splitter.count() > 1:
            action = QAction(
                "Show Control Area", self,
                objectName="action-show-control-area",
                shortcut=QKeySequence("Ctrl+Shift+D"),
                checkable=True,
                autoRepeat=False,
            )
            action.setChecked(True)
            action.triggered[bool].connect(self.__setControlAreaVisible)
            self.__splitter.handleClicked.connect(self.__toggleControlArea)
            viewaction.menu().addAction(action)

        if self.__menubar_action is not None:
            viewaction.menu().addAction(self.__menubar_action)

        if self.controlArea is not None:
            # Otherwise, the first control has focus
            self.controlArea.setFocus(Qt.OtherFocusReason)
        return self

    def menuBar(self) -> QMenuBar:
        return self.__menubar

    def __menuBarVisibleTimeout(self):
        mb = self.__menubar
        if mb is not None and mb.isHidden() \
                and QApplication.mouseButtons() == Qt.NoButton:
            mb.setVisible(True)
            mb.setProperty("__visible_from_alt_key_press", True)

    def __setMenuBarVisible(self, visible):
        mb = self.__menubar
        if mb is not None:
            mb.setVisible(visible)
            mb.setProperty("__visible_from_alt_key_press", False)
            settings = QSettings()
            settings.beginGroup(__name__ + "/menubar")
            settings.setValue("visible", visible)

    # pylint: disable=super-init-not-called
    def __init__(self, *args, **kwargs):
        """__init__s are called in __new__; don't call them from here"""

    def __init_subclass__(cls, **_):
        for base in cls.__bases__:
            if hasattr(base, "_final_class"):
                warnings.warn(
                    "subclassing of widget classes is deprecated and will be "
                    "disabled in the future.\n"
                    f"Extract code from {base.__name__} or explicitly open it "
                    "by adding `openclass=True` to class definition.",
                    RuntimeWarning, stacklevel=3)
                # raise TypeError(f"class {base.__name__} cannot be subclassed")

    @classmethod
    def get_widget_description(cls):
        if not cls.name:
            return None
        properties = {name: getattr(cls, name) for name in
                      ("name", "icon", "description", "priority", "keywords",
                       "replaces", "short_name", "category")}
        properties["id"] = cls.id or cls.__module__
        properties["inputs"] = cls.get_signals("inputs")
        properties["outputs"] = cls.get_signals("outputs")
        properties["qualified_name"] = f"{cls.__module__}.{cls.__name__}"
        return properties

    @classmethod
    def get_flags(cls):
        return Qt.Window if cls.resizing_enabled else Qt.Dialog

    class _Splitter(QSplitter):
        handleClicked = Signal()

        def __init__(self, *args, **kwargs):
            super().__init__(*args, *kwargs)
            self.setHandleWidth(18)

        def _adjusted_size(self, size_method):
            parent = self.parentWidget()
            if isinstance(parent, OWBaseWidget) \
                    and not parent.controlAreaVisible \
                    and self.count() > 1:
                indices = range(1, self.count())
            else:
                indices = range(0, self.count())
            shs = [size_method(self.widget(i))() for i in indices]
            height = max((sh.height() for sh in shs), default=0)
            width = sum(sh.width() for sh in shs)
            width += max(0, self.handleWidth() * (self.count() - 1))
            return QSize(width, height)

        def sizeHint(self):
            return self._adjusted_size(attrgetter("sizeHint"))

        def minimumSizeHint(self):
            return self._adjusted_size(attrgetter("minimumSizeHint"))

        def setSizes(self, sizes):
            super().setSizes(sizes)
            if len(sizes) == 2:
                self.handle(1).setControlAreaShown(bool(sizes[0]))

        def createHandle(self):
            """Create splitter handle"""
            return self._Handle(
                self.orientation(), self, cursor=Qt.PointingHandCursor)

        class _Handle(QSplitterHandle):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.controlAreaShown = True

            def setControlAreaShown(self, shown):
                self.controlAreaShown = shown
                self.update()

            def paintEvent(self, event):
                super(QSplitterHandle, self).paintEvent(event)
                painter = QPainter(self)
                pen = QPen()
                pen.setColor(QColor(160, 160, 160))
                pen.setCapStyle(Qt.RoundCap)
                pen.setWidth(3)
                painter.setPen(pen)
                w = self.width() - 6
                if self.controlAreaShown:
                    x0, x1 = 6, w
                else:
                    x0, x1 = w, 6
                y = self.height() // 2
                h = int((w - 6) / 1.12)
                painter.setRenderHint(painter.Antialiasing)
                painter.drawLines(
                    QLine(x0, y - h, x1, y),
                    QLine(x1, y, x0, y + h)
                )

            def mouseReleaseEvent(self, event):
                """Resize on left button"""
                if event.button() == Qt.LeftButton:
                    self.splitter().handleClicked.emit()
                super().mouseReleaseEvent(event)

            def mouseMoveEvent(self, event):
                """Prevent moving; just show/hide"""
                return

    def _insert_splitter(self):
        self.__splitter = self._Splitter(Qt.Horizontal, self)
        self.layout().addWidget(self.__splitter)

    def _insert_control_area(self):
        self.left_side = gui.vBox(self.__splitter, spacing=0)
        if self.want_main_area:
            self.left_side.setSizePolicy(
                QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

            scroll_area = VerticalScrollArea(self.left_side)
            scroll_area.setSizePolicy(QSizePolicy.MinimumExpanding,
                                      QSizePolicy.Preferred)
            self.controlArea = gui.vBox(scroll_area, spacing=6,
                                        sizePolicy=(QSizePolicy.MinimumExpanding,
                                                    QSizePolicy.Preferred))
            scroll_area.setWidget(self.controlArea)

            self.left_side.layout().addWidget(scroll_area)

            m = 4, 4, 0, 4
        else:
            self.controlArea = gui.vBox(self.left_side, spacing=6)

            m = 4, 4, 4, 4

        if self.buttons_area_orientation is not None:
            self._insert_buttons_area()
            self.buttonsArea.layout().setContentsMargins(
                m[0] + 8, m[1], m[2] + 8, m[3]
            )
            # margins are nice on macOS with this
            m = m[0], m[1], m[2], m[3] - 2

        self.controlArea.layout().setContentsMargins(*m)

    def _insert_buttons_area(self):
        if not self.want_main_area:
            gui.separator(self.left_side)
        self.buttonsArea = gui.widgetBox(
            self.left_side, spacing=6,
            orientation=self.buttons_area_orientation,
            sizePolicy=(QSizePolicy.MinimumExpanding,
                        QSizePolicy.Maximum)
        )

    def _insert_main_area(self):
        self.mainArea = gui.vBox(
            self.__splitter, spacing=6,
            sizePolicy=QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        )
        self.__splitter.addWidget(self.mainArea)
        self.__splitter.setCollapsible(
            self.__splitter.indexOf(self.mainArea), False)
        if self.want_control_area:
            self.mainArea.layout().setContentsMargins(
                0, 4, 4, 4)
            self.__splitter.setSizes([1, QWIDGETSIZE_MAX])
        else:
            self.mainArea.layout().setContentsMargins(
                4, 4, 4, 4)

    def _create_default_buttons(self):
        # These buttons are inserted in buttons_area, if it exists
        # Otherwise it is up to the widget to add them to some layout
        if self.graph_name is not None:
            self.graphButton = QPushButton("&Save Image", autoDefault=False)
            self.graphButton.clicked.connect(self.save_graph)
        if hasattr(self, "send_report"):
            self.report_button = QPushButton("&Report", autoDefault=False)
            self.report_button.clicked.connect(self.show_report)

    def set_basic_layout(self):
        """Provide the basic widget layout

        Which parts are created is regulated by class attributes
        `want_main_area`, `want_control_area`, `want_message_bar` and
        `buttons_area_orientation`, the presence of method `send_report`
        and attribute `graph_name`.
        """
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(2, 2, 2, 2)

        if not self.resizing_enabled:
            self.layout().setSizeConstraint(QVBoxLayout.SetFixedSize)

        self._create_default_buttons()

        self._insert_splitter()
        if self.want_control_area:
            self._insert_control_area()
        if self.want_main_area:
            self._insert_main_area()

        if self.want_message_bar:
            # statusBar() handles 'want_message_bar', 'send_report'
            # 'graph_name' ...
            _ = self.statusBar()

    __progressBar = None  # type: Optional[QProgressBar]
    __statusbar = None    # type: Optional[QStatusBar]
    __statusbar_action = None  # type: Optional[QAction]

    def statusBar(self):
        # type: () -> QStatusBar
        """
        Return the widget's status bar.

        The status bar can be hidden/shown (`self.statusBar().setVisible()`).

        Note
        ----
        The status bar takes control of the widget's bottom margin
        (`contentsMargins`) to layout itself in the OWBaseWidget.
        """
        statusbar = self.__statusbar

        if statusbar is None:
            # Use a OverlayWidget for status bar positioning.
            c = OverlayWidget(self, alignment=Qt.AlignBottom)
            c.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
            c.setWidget(self)
            c.setLayout(QVBoxLayout())
            c.layout().setContentsMargins(0, 0, 0, 0)
            self.__statusbar = statusbar = _StatusBar(
                c, objectName="owwidget-status-bar"
            )
            statusbar.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Maximum)

            if self.resizing_enabled:
                statusbar.setSizeGripEnabled(True)
            else:
                statusbar.setSizeGripEnabled(False)
                statusbar.setContentsMargins(0, 0, 7, 0)

            statusbar.ensurePolished()
            c.layout().addWidget(statusbar)

            # Reserve the bottom margins for the status bar
            margins = self.contentsMargins()
            margins.setBottom(statusbar.sizeHint().height())
            self.setContentsMargins(margins)
            statusbar.change.connect(self.__updateStatusBarOnChange)

            # Toggle status bar visibility. This action is not visible and
            # enabled by default. Client classes can inspect self.actions
            # and enable it if necessary.
            self.__statusbar_action = statusbar_action = QAction(
                "Show Status Bar", self, objectName="action-show-status-bar",
                toolTip="Show status bar", checkable=True,
                enabled=False, visible=False,
                shortcut=QKeySequence("Shift+Ctrl+\\")
            )
            if self.want_message_bar:
                self.message_bar = MessagesWidget(
                    defaultStyleSheet=(
                        "div.field-text { white-space: pre; }\n"
                        "div.field-detailed-text {\n"
                        "    margin-top: 0.5em; margin-bottom: 0.5em; \n"
                        "    margin-left: 1em; margin-right: 1em;\n"
                        "}"
                    ),
                    elideText=True,
                    sizePolicy=QSizePolicy(QSizePolicy.Preferred,
                                           QSizePolicy.Preferred),
                    visible=False
                )
            self.__progressBar = pb = QProgressBar(
                maximumWidth=120, minimum=0, maximum=100
            )
            pb.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Ignored)
            pb.setAttribute(Qt.WA_LayoutUsesWidgetRect)
            pb.setAttribute(Qt.WA_MacMiniSize)
            pb.hide()
            self.processingStateChanged.connect(self.__processingStateChanged)
            self.blockingStateChanged.connect(self.__processingStateChanged)
            self.progressBarValueChanged.connect(lambda v: pb.setValue(int(v)))

            statusbar.addPermanentWidget(pb)
            if self.message_bar is not None:
                statusbar.addPermanentWidget(self.message_bar)

            statusbar_action.toggled[bool].connect(statusbar.setVisible)
            self.addAction(statusbar_action)

            # reserve buttons and in_out_msg areas
            def hlayout(spacing, left=0, right=0, ):
                lay = QHBoxLayout(spacing=spacing)
                lay.setContentsMargins(left, 0, right, 0)
                return lay

            buttons = QWidget(statusbar, objectName="buttons", visible=False)
            buttons.setLayout(hlayout(5, 7))
            buttonsLayout = buttons.layout()
            simple_button = _StatusBar.simple_button
            icon = _load_styled_icon
            if self.__menubar is not None \
                    and not self.__menubar.isNativeMenuBar():
                # damn millennials
                b = _StatusBarButton(
                    icon=icon("hamburger.svg"),
                    toolTip="Menu",
                    objectName="status-bar-menu-button"
                )
                buttonsLayout.addWidget(b)
                b.clicked.connect(self.__showStatusBarMenu)

            if self.__help_action is not None:
                b = simple_button(buttons, self.__help_action, icon("help.svg"))
                buttonsLayout.addWidget(b)
            if self.__save_image_action is not None:
                b = simple_button(buttons, self.__save_image_action, icon("chart.svg"))
                buttonsLayout.addWidget(b)
            if self.__report_action is not None:
                b = simple_button(buttons, self.__report_action, icon("report.svg"))
                buttonsLayout.addWidget(b)
            if self.__reset_action is not None:
                b = simple_button(buttons, self.__reset_action, icon("reset.svg"))
                buttonsLayout.addWidget(b)
            if self.__visual_settings_action is not None:
                b = simple_button(buttons, self.__visual_settings_action, icon("visual-settings.svg"))
                buttonsLayout.addWidget(b)

            if buttonsLayout.count():
                buttons.setVisible(True)

            in_out_msg = QWidget(objectName="in-out-msg", visible=False)
            in_out_msg.setLayout(hlayout(5, left=5))
            statusbar.addWidget(buttons)
            statusbar.addWidget(in_out_msg)

            # Ensure the status bar and the message widget are visible on
            # warning and errors.
            self.messageActivated.connect(self.__ensureStatusBarVisible)

            if self.__menubar is not None:
                viewm = self.findChild(QMenu, "menu-view")
                if viewm is not None:
                    viewm.addAction(statusbar_action)

        return statusbar

    def __ensureStatusBarVisible(self, msg: Msg) -> None:
        statusbar = self.__statusbar
        if statusbar is not None and msg.group.severity >= 1:
            statusbar.setVisible(True)

    def __updateStatusBarOnChange(self):
        statusbar = self.__statusbar
        visible = statusbar.isVisibleTo(self)
        if visible:
            height = statusbar.height()
        else:
            height = 0
        margins = self.contentsMargins()
        margins.setBottom(height)
        self.setContentsMargins(margins)
        self.__statusbar_action.setChecked(visible)

    def __processingStateChanged(self):
        # Update the progress bar in the widget's status bar
        pb = self.__progressBar
        if pb is None:
            return
        pb.setVisible(bool(self.processingState) or self.isBlocking())
        if self.isBlocking() and not self.processingState:
            pb.setRange(0, 0)  # indeterminate pb
        elif self.processingState:
            pb.setRange(0, 100)  # determinate pb

    __info_ns = None  # type: Optional[StateInfo]

    def __info(self):
        # Create and return the StateInfo object
        if self.__info_ns is None:
            self.__info_ns = info = StateInfo(self)
            # default css for IO summary.
            css = textwrap.dedent("""
            /* vertical row header cell */
            tr > th.field-name {
                text-align: right;
                padding-right: 0.2em;
                font-weight: bold;
            }
            dt {
                font-weight: bold;
            }
            """)

            sb = self.statusBar()
            if sb is not None:
                in_out_msg = sb.findChild(QWidget, "in-out-msg")
                assert in_out_msg is not None
                in_out_msg.setVisible(True)
                in_msg = InOutStateWidget(
                    objectName="input-summary", visible=False,
                    defaultStyleSheet=css,
                    sizePolicy=QSizePolicy(QSizePolicy.Fixed,
                                           QSizePolicy.Fixed)
                )
                out_msg = InOutStateWidget(
                    objectName="output-summary", visible=False,
                    defaultStyleSheet=css,
                    sizePolicy=QSizePolicy(QSizePolicy.Fixed,
                                           QSizePolicy.Fixed)
                )
                in_msg.clicked.connect(partial(self.show_preview, self.input_summaries))
                out_msg.clicked.connect(partial(self.show_preview, self.output_summaries))

                # Insert a separator if these are not the first elements
                buttons = sb.findChild(QWidget, "buttons")
                assert buttons is not None
                if buttons.layout().count() != 0:
                    sep = QFrame(frameShape=QFrame.VLine)
                    sep.setContentsMargins(0, 0, 2, 0)
                    in_out_msg.layout().addWidget(sep)

                in_out_msg.layout().addWidget(in_msg)
                in_out_msg.layout().addWidget(out_msg)
                in_out_msg.setVisible(True)

                def set_message(msgwidget, m):
                    # type: (MessagesWidget, StateInfo.Summary) -> None
                    message = MessagesWidget.Message(
                        icon=m.icon, text=m.brief, informativeText=m.details,
                        textFormat=m.format
                    )
                    msgwidget.setMessage(message)
                    msgwidget.setVisible(not message.isEmpty())

                info.input_summary_changed.connect(
                    lambda m: set_message(in_msg, m)
                )
                info.output_summary_changed.connect(
                    lambda m: set_message(out_msg, m)
                )
        else:
            info = self.__info_ns
        return info

    @property
    def info(self):
        # type: () -> StateInfo
        """
        A namespace for reporting I/O, state ... related messages.

        .. versionadded:: 3.19

        Returns
        -------
        namespace : StateInfo
        """
        # back-compatibility; subclasses were free to assign self.info =
        # to any value. Preserve this.
        try:
            return self.__dict__["info"]
        except KeyError:
            pass
        return self.__info()

    @info.setter
    def info(self, val):
        warnings.warn(
            "'OWBaseWidget.info' is a property since 3.19 and will be made read "
            "only in v4.0.",
            DeprecationWarning, stacklevel=3
        )
        self.__dict__["info"] = val

    def __toggleControlArea(self):
        if self.__splitter is None or self.__splitter.count() < 2:
            return
        self.__setControlAreaVisible(not self.__splitter.sizes()[0])

    def __setControlAreaVisible(self, visible):
        # type: (bool) -> None
        if self.__splitter is None or self.__splitter.count() < 2:
            return
        self.controlAreaVisible = visible
        action = self.findChild(QAction, "action-show-control-area")
        if action is not None:
            action.setChecked(visible)
        splitter = self.__splitter  # type: QSplitter
        w = splitter.widget(0)
        # Set minimum width to 1 (overrides minimumSizeHint) when control area
        # is not visible to allow the main area to shrink further. Reset the
        # minimum width with a 0 if control area is visible.
        w.setMinimumWidth(int(not visible))

        sizes = splitter.sizes()
        current_size = sizes[0]
        if bool(current_size) == visible:
            return

        current_width = w.width()
        geom = self.geometry()
        frame = self.frameGeometry()
        framemargins = QMargins(
            frame.left() - geom.left(),
            frame.top() - geom.top(),
            frame.right() - geom.right(),
            frame.bottom() - geom.bottom()
        )
        splitter.setSizes([int(visible), QWIDGETSIZE_MAX])
        if not self.isWindow() or \
                self.windowState() not in [Qt.WindowNoState, Qt.WindowActive]:
            # not a window or not in state where we can move move/resize
            return

        # force immediate resize recalculation
        splitter.refresh()
        self.layout().invalidate()
        self.layout().activate()

        if visible:
            # move left and expand by the exposing widget's width
            diffx = -w.width()
            diffw = w.width()
        else:
            # move right and shrink by the collapsing width
            diffx = current_width
            diffw = -current_width
        newgeom = QRect(
            geom.x() + diffx, geom.y(), geom.width() + diffw, geom.height()
        )
        # bound/move by available geometry
        bounds = self.screen().availableGeometry()
        bounds = bounds.adjusted(
            framemargins.left(), framemargins.top(),
            -framemargins.right(), -framemargins.bottom()
        )
        newsize = newgeom.size().boundedTo(bounds.size())
        newgeom = QRect(newgeom.topLeft(), newsize)
        newgeom.moveLeft(max(newgeom.left(), bounds.left()))
        newgeom.moveRight(min(newgeom.right(), bounds.right()))
        self.setGeometry(newgeom)

    def save_graph(self):
        """Save the graph with the name given in class attribute `graph_name`.

        The method is called by the *Save graph* button, which is created
        automatically if the `graph_name` is defined.
        """
        graph_obj = getdeepattr(self, self.graph_name, None)
        if graph_obj is None:
            return
        saveplot.save_plot(graph_obj, self.graph_writers)

    def copy_to_clipboard(self):

        if self.graph_name:
            graph_obj = getdeepattr(self, self.graph_name, None)
            if graph_obj is None:
                return
            ClipboardFormat.write_image(None, graph_obj)

    def __restoreWidgetGeometry(self, geometry):
        # type: (bytes) -> bool
        def _fullscreen_to_maximized(geometry):
            """Don't restore windows into full screen mode because it loses
            decorations and can't be de-fullscreened at least on some platforms.
            Use Maximized state insted."""
            w = QWidget(visible=False)
            w.restoreGeometry(QByteArray(geometry))
            if w.isFullScreen():
                w.setWindowState(
                    w.windowState() & ~Qt.WindowFullScreen | Qt.WindowMaximized)
            return w.saveGeometry()

        restored = False
        if geometry:
            geometry = _fullscreen_to_maximized(geometry)
            restored = self.restoreGeometry(geometry)

        if restored and not self.windowState() & \
                (Qt.WindowMaximized | Qt.WindowFullScreen):
            space = self.screen().availableGeometry()
            frame, geometry = self.frameGeometry(), self.geometry()

            # Fix the widget size to fit inside the available space
            width = space.width() - (frame.width() - geometry.width())
            width = min(width, geometry.width())
            height = space.height() - (frame.height() - geometry.height())
            height = min(height, geometry.height())
            self.resize(width, height)

            # Move the widget to the center of available space if it is
            # currently outside it
            if not space.contains(self.frameGeometry()):
                x = max(0, space.width() // 2 - width // 2)
                y = max(0, space.height() // 2 - height // 2)

                self.move(x, y)
        return restored

    def __updateSavedGeometry(self):
        if self.__was_restored and self.isVisible():
            # Update the saved geometry only between explicit show/hide
            # events (i.e. changes initiated by the user not by Qt's default
            # window management).
            # Note: This should always be stored as bytes and not QByteArray.
            self.savedWidgetGeometry = bytes(self.saveGeometry())

    def sizeHint(self):
        if not self.want_basic_layout \
                or self.mainArea_width_height_ratio is None:
            return super().sizeHint()

        # Super sizeHint with scroll_area isn't calculated right (slightly too small on macOS)
        # on some platforms. This way, width/height should be optimal for most widgets.
        sh = QSize()
        # boxes look nice on macOS with starting width/height 4
        width = 4
        height = 4

        if self.want_message_bar:
            msh = self.statusBar().sizeHint()
            height += msh.height()
        if self.want_control_area:
            csh = self.controlArea.sizeHint()
            width += csh.width()
            height += csh.height()
            if self.buttons_area_orientation:
                bsh = self.buttonsArea.sizeHint()
                height += bsh.height()
        height = max(height, 500)
        if self.want_main_area:
            width += self.__splitter.handleWidth()
            if self.want_control_area:
                width += int(height * self.mainArea_width_height_ratio)
            else:
                return super().sizeHint()
        return QSize(width, height)

    # when widget is resized, save the new width and height
    def resizeEvent(self, event):
        """Overloaded to save the geometry (width and height) when the widget
        is resized.
        """
        QDialog.resizeEvent(self, event)
        # Don't store geometry if the widget is not visible
        # (the widget receives a resizeEvent (with the default sizeHint)
        # before first showEvent and we must not overwrite the the
        # savedGeometry with it)
        if self.save_position and self.isVisible():
            self.__updateSavedGeometry()

    def moveEvent(self, event):
        """Overloaded to save the geometry when the widget is moved
        """
        QDialog.moveEvent(self, event)
        if self.save_position and self.isVisible():
            self.__updateSavedGeometry()

    def hideEvent(self, event):
        """Overloaded to save the geometry when the widget is hidden
        """
        if self.save_position:
            self.__updateSavedGeometry()
        QDialog.hideEvent(self, event)

    def closeEvent(self, event):
        """Overloaded to save the geometry when the widget is closed
        """
        if self.save_position and self.isVisible():
            self.__updateSavedGeometry()
        QDialog.closeEvent(self, event)

    def mousePressEvent(self, event):
        """ Flash message bar icon on mouse press """
        if self.message_bar is not None:
            self.message_bar.flashIcon()
        event.ignore()

    def setVisible(self, visible):
        # type: (bool) -> None
        """Reimplemented from `QDialog.setVisible`."""
        if visible:
            # Force cached size hint invalidation ... The size hints are
            # sometimes not properly invalidated via the splitter's layout and
            # nested left_part -> controlArea layouts. This causes bad initial
            # size when the widget is first shown.
            if self.controlArea is not None:
                self.controlArea.updateGeometry()
            if self.buttonsArea is not None:
                self.buttonsArea.updateGeometry()
            if self.mainArea is not None:
                self.mainArea.updateGeometry()
        super().setVisible(visible)

    def showEvent(self, event):
        """Overloaded to restore the geometry when the widget is shown
        """
        QDialog.showEvent(self, event)
        if not self.__was_restored:
            # Restore saved geometry/layout on (first) show
            if self.__splitter is not None:
                self.__setControlAreaVisible(self.controlAreaVisible)
            if self.save_position and self.savedWidgetGeometry is not None:
                self.__restoreWidgetGeometry(bytes(self.savedWidgetGeometry))
            self.__was_restored = True

        if not self.__was_shown:
            # Mark as explicitly moved/resized if not already. QDialog would
            # otherwise adjust position/size on subsequent hide/show
            # (move/resize events coming from the window manager do not set
            # these flags).
            self.setAttribute(Qt.WA_Moved, True)
            self.setAttribute(Qt.WA_Resized, True)
            self.__was_shown = True
        self.__quicktipOnce()

    def __showStatusBarMenu(self):
        # type: () -> None
        sb = self.__statusbar
        mb = self.__menubar
        if sb is None or mb is None:
            return
        b = sb.findChild(SimpleButton, "status-bar-menu-button")
        if b is None:
            return

        actions = []
        for action in mb.actions():
            if action.isVisible() and action.isEnabled() and action.menu():
                actions.append(action)
        if not actions:
            return
        menu = QMenu(self)
        menu.setAttribute(Qt.WA_DeleteOnClose)
        menu.addActions(actions)
        popup_rect = QRect(
            b.mapToGlobal(QPoint(0, 0)), b.size()
        )
        menu.ensurePolished()
        screen_rect = b.screen().availableGeometry()
        menu_rect = dropdown_popup_geometry(
            menu.sizeHint(), popup_rect, screen_rect, preferred_direction="up"
        )
        menu.popup(menu_rect.topLeft())

    def setCaption(self, caption):
        # save caption title in case progressbar will change it
        self.captionTitle = str(caption)
        self.setWindowTitle(caption)

    def reshow(self):
        """Put the widget on top of all windows
        """
        self.show()
        self.raise_()
        self.activateWindow()

    def openContext(self, *a):
        """Open a new context corresponding to the given data.

        The settings handler first checks the stored context for a
        suitable match. If one is found, it becomes the current contexts and
        the widgets settings are initialized accordingly. If no suitable
        context exists, a new context is created and data is copied from
        the widget's settings into the new context.

        Widgets that have context settings must call this method after
        reinitializing the user interface (e.g. combo boxes) with the new
        data.

        The arguments given to this method are passed to the context handler.
        Their type depends upon the handler. For instance,
        `DomainContextHandler` expects `Orange.data.Table` or
        `Orange.data.Domain`.
        """
        self.contextAboutToBeOpened.emit(a)
        self.settingsHandler.open_context(self, *a)
        self.contextOpened.emit()

    def closeContext(self):
        """Save the current settings and close the current context.

        Widgets that have context settings must call this method before
        reinitializing the user interface (e.g. combo boxes) with the new
        data.
        """
        self.settingsHandler.close_context(self)
        self.contextClosed.emit()

    def retrieveSpecificSettings(self):
        """
        Retrieve data that is not registered as setting.

        This method is called by
        `orangewidget.settings.ContextHandler.settings_to_widget`.
        Widgets may define it to retrieve any data that is not stored in widget
        attributes. See :obj:`Orange.widgets.data.owcolor.OWColor` for an
        example.
        """

    def storeSpecificSettings(self):
        """
        Store data that is not registered as setting.

        This method is called by
        `orangewidget.settings.ContextHandler.settings_from_widget`.
        Widgets may define it to store any data that is not stored in widget
        attributes. See :obj:`Orange.widgets.data.owcolor.OWColor` for an
        example.
        """

    def saveSettings(self):
        """
        Writes widget instance's settings to class defaults. Usually called
        when the widget is deleted.
        """
        self.settingsHandler.update_defaults(self)

    def onDeleteWidget(self):
        """
        Invoked by the canvas to notify the widget it has been deleted
        from the workflow.

        If possible, subclasses should gracefully cancel any currently
        executing tasks.
        """

    def handleNewSignals(self):
        """
        Invoked by the workflow signal propagation manager after all
        signals handlers have been called.

        Reimplement this method in order to coalesce updates from
        multiple updated inputs.
        """

    #: Widget's status message has changed.
    statusMessageChanged = Signal(str)

    def setStatusMessage(self, text):
        """
        Set widget's status message.

        This is a short status string to be displayed inline next to
        the instantiated widget icon in the canvas.
        """
        assert QThread.currentThread() == self.thread()
        if self.__statusMessage != text:
            self.__statusMessage = text
            self.statusMessageChanged.emit(text)

    def statusMessage(self):
        """
        Return the widget's status message.
        """
        return self.__statusMessage

    def keyPressEvent(self, e: QKeyEvent) -> None:
        mb = self.__menubar
        if not mb.isNativeMenuBar() \
                and e.modifiers() == Qt.AltModifier \
                and e.key() in [Qt.Key_Alt, Qt.Key_AltGr] \
                and QApplication.mouseButtons() == Qt.NoButton \
                and mb.isHidden():
            self.__menubar_visible_timer.start()
        elif self.__menubar_visible_timer is not None:
            # stop the timer on any other key press
            self.__menubar_visible_timer.stop()
        super().keyPressEvent(e)

    def keyReleaseEvent(self, event: QKeyEvent) -> None:
        mb = self.__menubar
        if not mb.isNativeMenuBar() \
                and event.key() in [Qt.Key_Alt, Qt.Key_AltGr]:
            self.__menubar_visible_timer.stop()
            if mb.property("__visible_from_alt_key_press") is True:
                mb.setVisible(False)
                mb.setProperty("__visible_from_alt_key_press", False)
        super().keyReleaseEvent(event)

    def setBlocking(self, state=True) -> None:
        """
        Set blocking flag for this widget.

        While this flag is set this widget and all its descendants
        will not receive any new signals from the workflow signal manager.

        .. deprecated:: 4.2.0
            Setting/clearing this flag is equivalent to
            `setInvalidated(True); setReady(False)` and
            `setInvalidated(False); setReady(True)` respectively.
            Use :func:`setInvalidated` and :func:`setReady` in new code.

        .. seealso:: :func:`setInvalidated`, :func:`setReady`

        """
        if state:
            self.__setState(True, False)
        else:
            self.__setState(False, True)

    def isBlocking(self):
        """
        Is this widget blocking signal processing.
        """
        return self.isInvalidated() and not self.isReady()

    widgetStateChanged = Signal()
    blockingStateChanged = Signal(bool)
    processingStateChanged = Signal(int)
    invalidatedStateChanged = Signal(bool)
    readyStateChanged = Signal(bool)

    __invalidated = False
    __ready = True

    def setInvalidated(self, state: bool) -> None:
        """
        Set/clear invalidated flag on this widget.

        While this flag is set none of its descendants will receive new
        signals from the workflow execution manager.

        This is useful for instance if the widget does it's work in a
        separate thread or schedules processing from the event queue.
        In this case it can set the invalidated flag when starting a task.
        After the task has completed the widget can clear the flag and
        send the updated outputs.

        .. note::
            Failure to clear this flag will block dependent nodes forever.

        .. seealso:: :func:`~Output.invalidate()` for a more fine grained
           invalidation.
        """
        self.__setState(state, self.__ready)

    def isInvalidated(self) -> bool:
        """
        Return the widget's invalidated flag.
        """
        return self.__invalidated

    def setReady(self, state: bool) -> None:
        """
        Set/clear ready flag on this widget.

        While a ready flag is unset, the widget will not receive any new
        input updates from the workflow execution manager.

        By default this flag is True.
        """
        self.__setState(self.__invalidated, state)

    def isReady(self) -> bool:
        """
        Return the widget's ready state
        """
        return self.__ready

    def __setState(self, invalidated: bool, ready: bool) -> None:
        blocking = self.isBlocking()
        changed = False
        if self.__ready != ready:
            self.__ready = ready
            changed = True
            self.readyStateChanged.emit(ready)
        if self.__invalidated != invalidated:
            self.__invalidated = invalidated
            self.invalidatedStateChanged.emit(invalidated)
            changed = True
        if changed:
            self.widgetStateChanged.emit()
        if blocking != self.isBlocking():
            self.blockingStateChanged.emit(self.isBlocking())

    def workflowEnv(self):
        """
        Return (a view to) the workflow runtime environment.

        Returns
        -------
        env : types.MappingProxyType
        """
        return self.__env

    def workflowEnvChanged(self, key, value, oldvalue):
        """
        A workflow environment variable `key` has changed to value.

        Called by the canvas framework to notify widget of a change
        in the workflow runtime environment.

        The default implementation does nothing.
        """

    def saveGeometryAndLayoutState(self):
        # type: () -> QByteArray
        """
        Save the current geometry and layout state of this widget and
        child windows (if applicable).

        Returns
        -------
        state : QByteArray
            Saved state.
        """
        version = 0x1
        have_spliter = 0
        splitter_state = 0
        if self.__splitter is not None:
            have_spliter = 1
            splitter_state = 1 if self.controlAreaVisible else 0
        data = QByteArray()
        stream = QDataStream(data, QBuffer.WriteOnly)
        stream.writeUInt32(version)
        stream.writeUInt16((have_spliter << 1) | splitter_state)
        stream <<= self.saveGeometry()
        return data

    def restoreGeometryAndLayoutState(self, state):
        # type: (QByteArray) -> bool
        """
        Restore the geometry and layout of this widget to a state previously
        saved with :func:`saveGeometryAndLayoutState`.

        Parameters
        ----------
        state : QByteArray
            Saved state.

        Returns
        -------
        success : bool
            `True` if the state was successfully restored, `False` otherwise.
        """
        version = 0x1
        stream = QDataStream(state, QBuffer.ReadOnly)
        version_ = stream.readUInt32()
        if stream.status() != QDataStream.Ok or version_ != version:
            return False
        splitter_state = stream.readUInt16()
        has_spliter = splitter_state & 0x2
        splitter_state = splitter_state & 0x1
        if has_spliter and self.__splitter is not None:
            self.__setControlAreaVisible(bool(splitter_state))
        geometry = QByteArray()
        stream >>= geometry
        if stream.status() == QDataStream.Ok:
            state = self.__restoreWidgetGeometry(bytes(geometry))
            self.__was_restored = self.__was_restored or state
            return state
        else:
            return False  # pragma: no cover

    def __showMessage(self, message):
        if self.__msgwidget is not None:
            self.__msgwidget.hide()
            self.__msgwidget.deleteLater()
            self.__msgwidget = None

        if message is None:
            return

        buttons = MessageOverlayWidget.Ok | MessageOverlayWidget.Close
        if message.moreurl is not None:
            buttons |= MessageOverlayWidget.Help

        if message.icon is not None:
            icon = message.icon
        else:
            icon = Message.Information

        self.__msgwidget = MessageOverlayWidget(
            parent=self, text=message.text, icon=icon, wordWrap=True,
            standardButtons=buttons)

        btn = self.__msgwidget.button(MessageOverlayWidget.Ok)
        btn.setText("Ok, got it")

        self.__msgwidget.setStyleSheet("""
            MessageOverlayWidget {
                background: qlineargradient(
                    x1: 0, y1: 0, x2: 0, y2: 1,
                    stop:0 #666, stop:0.3 #6D6D6D, stop:1 #666)
            }
            MessageOverlayWidget QLabel#text-label {
                color: white;
            }""")

        if message.moreurl is not None:
            helpbutton = self.__msgwidget.button(MessageOverlayWidget.Help)
            helpbutton.setText("Learn more\N{HORIZONTAL ELLIPSIS}")
            self.__msgwidget.helpRequested.connect(
                lambda: QDesktopServices.openUrl(QUrl(message.moreurl)))

        self.__msgwidget.setWidget(self)
        self.__msgwidget.show()

    def __quicktip(self):
        messages = list(self.UserAdviceMessages)
        if messages:
            message = messages[self.__msgchoice % len(messages)]
            self.__msgchoice += 1
            self.__showMessage(message)

    def __quicktipOnce(self):
        dirpath = settings.widget_settings_dir(versioned=False)
        try:
            os.makedirs(dirpath, exist_ok=True)
        except OSError:  # EPERM, EEXISTS, ...
            pass

        filename = os.path.join(settings.widget_settings_dir(versioned=False),
                                "user-session-state.ini")
        namespace = ("user-message-history/{0.__module__}.{0.__qualname__}"
                     .format(type(self)))
        session_hist = QSettings(filename, QSettings.IniFormat)
        session_hist.beginGroup(namespace)
        messages = self.UserAdviceMessages

        def _ispending(msg):
            return not session_hist.value(
                "{}/confirmed".format(msg.persistent_id),
                defaultValue=False, type=bool)
        messages = [msg for msg in messages if _ispending(msg)]

        if not messages:
            return

        message = messages[self.__msgchoice % len(messages)]
        self.__msgchoice += 1

        self.__showMessage(message)

        def _userconfirmed():
            session_hist = QSettings(filename, QSettings.IniFormat)
            session_hist.beginGroup(namespace)
            session_hist.setValue(
                "{}/confirmed".format(message.persistent_id), True)
            session_hist.sync()

        self.__msgwidget.accepted.connect(_userconfirmed)

    @classmethod
    def migrate_settings(cls, settings, version):
        """Fix settings to work with the current version of widgets

        Parameters
        ----------
        settings : dict
            dict of name - value mappings
        version : Optional[int]
            version of the saved settings
            or None if settings were created before migrations
        """

    @classmethod
    def migrate_context(cls, context, version):
        """Fix contexts to work with the current version of widgets

        Parameters
        ----------
        context : Context
            Context object
        version : Optional[int]
            version of the saved context
            or None if context was created before migrations
        """

    def actionEvent(self, event: QActionEvent) -> None:
        if event.type() in (QEvent.ActionAdded, QEvent.ActionRemoved):
            event = cast(QActionEvent, event)
            action = event.action()
            if action.objectName().startswith("action-canvas-"):
                menu = self.findChild(QMenu, "menu-window")
                if menu is not None:
                    if event.type() == QEvent.ActionAdded:
                        menu.addAction(action)
                    else:
                        menu.removeAction(action)
        super().actionEvent(event)


class _StatusBar(QStatusBar):
    #: Emitted on a change of geometry or visibility (explicit hide/show)
    change = Signal()

    def event(self, event):
        # type: (QEvent) ->bool
        if event.type() in {QEvent.Resize, QEvent.ShowToParent,
                            QEvent.HideToParent}:
            self.change.emit()
        return super().event(event)

    def paintEvent(self, event):
        style = self.style()
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        # Omit the widget instance from the call (QTBUG-60018)
        style.drawPrimitive(QStyle.PE_PanelStatusBar, opt, painter, None)
        # Do not draw any PE_FrameStatusBarItem frames.
        painter.end()

    @staticmethod
    def simple_button(
            parent: QWidget, action: QAction, icon=QIcon()
    ) -> SimpleButton:
        if icon.isNull():
            icon = action.icon()
        button = _StatusBarButton(
            parent,
            icon=icon,
            toolTip=action.toolTip(), whatsThis=action.whatsThis(),
            visible=action.isVisible(), enabled=action.isEnabled(), )

        def update():
            button.setVisible(action.isVisible())
            button.setEnabled(action.isEnabled())
            button.setToolTip(action.toolTip())
            button.setWhatsThis(action.whatsThis())

        action.changed.connect(update)
        button.clicked.connect(action.triggered)
        return button


class _StatusBarButton(SimpleButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # match top/bottom margins of MessagesWidget
        self.setContentsMargins(1, 1, 1, 1)
        self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)

    def sizeHint(self):
        # Ensure the button has at least font height dimensions.
        sh = super().sizeHint()
        h = self.fontMetrics().lineSpacing()
        return sh.expandedTo(QSize(h, h))


class _Menu(QMenu):
    """
    A QMenu managing self-visibility in a parent menu or menu bar.

    The menu is visible if it has at least one visible action.
    """
    def actionEvent(self, event):
        super().actionEvent(event)
        ma = self.menuAction()
        if ma is not None:
            ma.setVisible(
                any(ac.isVisible() and not ac.isSeparator()
                    for ac in self.actions())
            )


#: Input/Output flags (deprecated).
#: --------------------------------
#:
#: The input/output is the default for its type.
#: When there are multiple IO signals with the same type the
#: one with the default flag takes precedence when adding a new
#: link in the canvas.
Default = Default
NonDefault = NonDefault
#: Single input signal (default)
Single = Single
#: Multiple outputs can be linked to this signal.
#: Signal handlers with this flag have (object, id: object) -> None signature.
Multiple = Multiple
#: Applies to user interaction only.
#: Only connected if specifically requested (in a dedicated "Links" dialog)
#: or it is the only possible connection.
Explicit = Explicit
#: Dynamic output type.
#: Specifies that the instances on the output will in general be
#: subtypes of the declared type and that the output can be connected
#: to any input signal which can accept a subtype of the declared output
#: type.
Dynamic = Dynamic


metric_suffix = ['', 'k', 'M', 'G', 'T', 'P']


class StateInfo(QObject):
    """
    A namespace for OWBaseWidget's detailed input/output/state summary reporting.

    See Also
    --------
    OWBaseWidget.info
    """
    class Summary:
        """
        Input/output summary description.

        This class is used to hold and report detailed I/O summaries.

        Attributes
        ----------
        brief: str
            A brief (inline) description.
        details: str
            A richer detailed description.
        icon: QIcon
            An custom icon. If not set a default set will be used to indicate
            special states (i.e. empty input ...)
        format: Qt.TextFormat
            Qt.PlainText if `brief` and `details` are to be rendered as plain
            text or Qt.RichText if they are HTML.

        See also
        --------
        :func:`StateInfo.set_input_summary`,
        :func:`StateInfo.set_output_summary`,
        :class:`StateInfo.Empty`,
        :class:`StateInfo.Partial`,
        `Supported HTML Subset`_

        .. _`Supported HTML Subset`:
            http://doc.qt.io/qt-5/richtext-html-subset.html

        """
        def __init__(self, brief="", details="", icon=QIcon(),
                     format=Qt.PlainText):
            # type: (str, str, QIcon, Qt.TextFormat) -> None
            super().__init__()
            self.__brief = brief
            self.__details = details
            self.__icon = QIcon(icon)
            self.__format = format

        @property
        def brief(self) -> str:
            return self.__brief

        @property
        def details(self) -> str:
            return self.__details

        @property
        def icon(self) -> QIcon:
            return QIcon(self.__icon)

        @property
        def format(self) -> Qt.TextFormat:
            return self.__format

        def __eq__(self, other):
            return (isinstance(other, StateInfo.Summary) and
                    self.brief == other.brief and
                    self.details == other.details and
                    self.icon.cacheKey() == other.icon.cacheKey() and
                    self.format == other.format)

        def as_dict(self):
            return dict(brief=self.brief, details=self.details, icon=self.icon,
                        format=self.format)

        def updated(self, **kwargs):
            state = self.as_dict()
            state.update(**kwargs)
            return type(self)(**state)

        @classmethod
        def default_icon(cls, role):
            # type: (str) -> QIcon
            """
            Return a default icon for input/output role.

            Parameters
            ----------
            role: str
                "input" or "output"

            Returns
            -------
            icon: QIcon
            """
            return _load_styled_icon(f"{role}.svg")

    class Empty(Summary):
        """
        Input/output summary description indicating empty I/O state.
        """
        @classmethod
        def default_icon(cls, role):
            return _load_styled_icon(f"{role}-empty.svg")

    class Partial(Summary):
        """
        Input summary indicating partial input.

        This state indicates that some inputs are present but more are needed
        in order for the widget to proceed.
        """
        @classmethod
        def default_icon(cls, role):
            return _load_styled_icon(f"{role}-partial.svg")

    #: Signal emitted when the input summary changes
    input_summary_changed = Signal(Summary)
    #: Signal emitted when the output summary changes
    output_summary_changed = Signal(Summary)

    #: A default message displayed to indicate no inputs.
    NoInput = Empty()

    #: A default message displayed to indicate no output.
    NoOutput = Empty()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__input_summary = StateInfo.Summary()   # type: StateInfo.Summary
        self.__output_summary = StateInfo.Summary()  # type: StateInfo.Summary

    def set_input_summary(self, summary, details="", icon=QIcon(),
                          format=Qt.PlainText):
        # type: (Union[StateInfo.Summary, str, int, None], str, QIcon, Qt.TextFormat) -> None
        """
        Set the input summary description.

        This method has two overloads

        .. function:: set_input_summary(summary: Optional[StateInfo.Summary]])

        .. function:: set_input_summary(summary:str, detailed:str="", icon:QIcon)

        Note
        ----
        `set_input_summary(None)` clears/resets the current summary. Use
        `set_input_summary(StateInfo.NoInput)` to indicate no input state.
        `set_input_summary(int)` to have automatically formatted summary

        Parameters
        ----------
        summary : Union[Optional[StateInfo.Message], str, int]
            A populated `StateInfo.Message` instance or a short text
            description (should not exceed 16 characters) or an integer.
        details : str
            A detailed description (only applicable if summary is a string).
        icon : QIcon
            An icon. If not specified a default icon will be used (only
            applicable if `summary` is a string).
        format : Qt.TextFormat
            Specify how the `short` and `details` text should be interpreted.
            Can be `Qt.PlainText` or `Qt.RichText` (only applicable if
            `summary` is a string or an integer).
        """
        def assert_single_arg():
            if not (details == "" and icon.isNull() and format == Qt.PlainText):
                raise TypeError("No extra arguments expected when `summary` "
                                "is `None` or `Message`")

        if summary is None:
            assert_single_arg()
            summary = StateInfo.Summary()
        elif isinstance(summary, StateInfo.Summary):
            assert_single_arg()
            if isinstance(summary, StateInfo.Empty):
                summary = summary.updated(details="No data on input",
                                          brief='-')
            if summary.icon.isNull():
                summary = summary.updated(icon=summary.default_icon("input"))
        elif isinstance(summary, str):
            summary = StateInfo.Summary(summary, details, icon, format=format)
            if summary.icon.isNull():
                summary = summary.updated(icon=summary.default_icon("input"))
        elif isinstance(summary, int):
            summary = StateInfo.Summary(self.format_number(summary),
                                        details or str(summary),
                                        StateInfo.Summary.default_icon("input"),
                                        format=format)
        else:
            raise TypeError("'None', 'str' or 'Message' instance expected, "
                            "got '{}'" .format(type(summary).__name__))

        if self.__input_summary != summary:
            self.__input_summary = summary
            self.input_summary_changed.emit(summary)

    def set_output_summary(self, summary, details="", icon=QIcon(),
                           format=Qt.PlainText):
        # type: (Union[StateInfo.Summary, str, int, None], str, QIcon, Qt.TextFormat) -> None
        """
        Set the output summary description.

        Note
        ----
        `set_output_summary(None)` clears/resets the current summary. Use
        `set_output_summary(StateInfo.NoOutput)` to indicate no output state.
        `set_output_summary(int)` to have automatically formatted summary

        Parameters
        ----------
        summary : Union[StateInfo.Summary, str, int, None]
            A populated `StateInfo.Summary` instance or a short text
            description (should not exceed 16 characters) or an integer.
        details : str
            A detailed description (only applicable if `summary` is a string).
        icon : QIcon
            An icon. If not specified a default icon will be used
            (only applicable if `summary` is a string)
        format : Qt.TextFormat
            Specify how the `summary` and `details` text should be interpreted.
            Can be `Qt.PlainText` or `Qt.RichText` (only applicable if
            `summary` is a string or an integer).
        """
        def assert_single_arg():
            if not (details == "" and icon.isNull() and format == Qt.PlainText):
                raise TypeError("No extra arguments expected when `summary` "
                                "is `None` or `Message`")
        if summary is None:
            assert_single_arg()
            summary = StateInfo.Summary()
        elif isinstance(summary, StateInfo.Summary):
            assert_single_arg()
            if isinstance(summary, StateInfo.Empty):
                summary = summary.updated(details="No data on output",
                                          brief='-')
            if summary.icon.isNull():
                summary = summary.updated(icon=summary.default_icon("output"))
        elif isinstance(summary, str):
            summary = StateInfo.Summary(summary, details, icon, format=format)
            if summary.icon.isNull():
                summary = summary.updated(icon=summary.default_icon("output"))
        elif isinstance(summary, int):
            summary = StateInfo.Summary(self.format_number(summary),
                                        details or str(summary),
                                        StateInfo.Summary.default_icon("output"),
                                        format=format)
        else:
            raise TypeError("'None', 'str' or 'Message' instance expected, "
                            "got '{}'" .format(type(summary).__name__))

        if self.__output_summary != summary:
            self.__output_summary = summary
            self.output_summary_changed.emit(summary)

    @staticmethod
    def format_number(n: int) -> str:
        """
        Format integers larger then 9999 with metric suffix and at most 3 digits.

        Example:
            >>> StateInfo.format_number(12_345)
            '12.3k'
        """
        if n < 10_000:
            return str(n)
        mag = int(log10(n) // 3)
        n = n / 10 ** (mag * 3)
        if n >= 999.5:
            # rounding to higher order
            n = 1
            mag += 1
        return f"{n:.3g}{metric_suffix[mag]}"

# pylint: disable=too-many-lines



import enum
import inspect
from typing import Union, Iterator, Optional

import sys
import warnings
from operator import attrgetter

from AnyQt.QtCore import QObject, QRect, QSize, QPoint, QTextBoundaryFinder


def progress_bar_milestones(count, iterations=100):
    return set([int(i*count/float(iterations)) for i in range(iterations)])


_NOTSET = object()


def deepgetattr(obj, attr, default=_NOTSET):
    """Works exactly like getattr(), except that attr can be a nested attribute
    (e.g. "attr1.attr2.attr3").
    """
    try:
        return attrgetter(attr)(obj)
    except AttributeError:
        if default is _NOTSET:
            raise
        return default


def getdeepattr(obj, attr, *arg, **kwarg):
    if isinstance(obj, dict):
        return obj.get(attr)
    return deepgetattr(obj, attr, *arg, **kwarg)


def to_html(str):
    return str.replace("<=", "&#8804;").replace(">=", "&#8805;").\
        replace("<", "&#60;").replace(">", "&#62;").replace("=\\=", "&#8800;")

getHtmlCompatibleString = to_html


def dumpObjectTree(obj, _indent=0):
    """
    Dumps Qt QObject tree. Aids in debugging internals.
    See also: QObject.dumpObjectTree()
    """
    assert isinstance(obj, QObject)
    print('{indent}{type} "{name}"'.format(indent=' ' * (_indent * 4),
                                           type=type(obj).__name__,
                                           name=obj.objectName()),
          file=sys.stderr)
    for child in obj.children():
        dumpObjectTree(child, _indent + 1)


def getmembers(obj, predicate=None):
    """Return all the members of an object in a list of (name, value) pairs sorted by name.

    Behaves like inspect.getmembers. If a type object is passed as a predicate,
    only members of that type are returned.
    """

    if isinstance(predicate, type):
        def mypredicate(x):
            return isinstance(x, predicate)
    else:
        mypredicate = predicate
    return inspect.getmembers(obj, mypredicate)


class DeprecatedSignal:
    def __init__(self, actual_signal, *args,
                 warning_text='Deprecated', emit_callback=None, **kwargs):
        self.signal = actual_signal
        self.warning_text = warning_text
        self.emit_callback = emit_callback

    def emit(self, *args, **kwargs):
        warnings.warn(
            self.warning_text,
            DeprecationWarning, stacklevel=2
        )
        if self.emit_callback:
            self.emit_callback(*args, **kwargs)
        return self.signal.emit(*args, **kwargs)

    def __getattr__(self, item):
        return self.__signal.item


def enum_as_int(value: Union[int, enum.Enum]) -> int:
    """
    Return a `enum.Enum` value as an `int.

    This is function intended for extracting underlying Qt5/6 enum
    values specifically with PyQt6 where most Qt enums are represented
    with `enum.Enum` and lose their numerical value.

    >>> from PyQt6.QtCore import Qt
    >>> enum_as_int(Qt.Alignment.AlignLeft)
    1
    """
    if isinstance(value, enum.Enum):
        return int(value.value)
    else:
        return int(value)


def dropdown_popup_geometry(
        size: QSize, origin: QRect, screen: QRect, preferred_direction="down"
) -> QRect:
    """
    Move/constrain the geometry for a drop down popup.

    Parameters
    ----------
    size : QSize
        The base popup size if not constrained.
    origin : QRect
        The origin rect from which the popup extends (in screen coords.).
    screen : QRect
        The available screen geometry into which the popup must fit.
    preferred_direction : str
        'up' or 'down'

    Returns
    -------
    geometry: QRect
        Constrained drop down list geometry to fit into screen
    """
    if preferred_direction == "down":
        # if the popup  geometry extends bellow the screen and there is more
        # room above the popup origin ...
        geometry = QRect(origin.bottomLeft() + QPoint(0, 1), size)
        if geometry.bottom() > screen.bottom() \
                and origin.center().y() > screen.center().y():
            # ...flip the rect about the origin so it extends upwards
            geometry.moveBottom(origin.top() - 1)
    elif preferred_direction == "up":
        geometry = QRect(origin.topLeft() - QPoint(0, 1 + size.height()), size)
        if geometry.top() < screen.top() \
                and origin.center().y() < screen.center().y():
            # ... flip, extend down
            geometry.moveTop(origin.bottom() - 1)
    else:
        raise ValueError(f"Invalid 'preferred_direction' ({preferred_direction})")

    # fixup horizontal position if it extends outside the screen
    if geometry.left() < screen.left():
        geometry.moveLeft(screen.left())
    if geometry.right() > screen.right():
        geometry.moveRight(screen.right())

    # bounded by screen geometry
    return geometry.intersected(screen)


def graphemes(text: str) -> Iterator[str]:
    """
    Return an iterator over grapheme clusters of text
    """
    # match internal QString encoding
    text_encoded = text.encode("utf-16-le")
    finder = QTextBoundaryFinder(QTextBoundaryFinder.Grapheme, text)
    start = 0
    while True:
        pos = finder.toNextBoundary()
        if pos == -1:
            return
        yield text_encoded[start*2: pos*2].decode("utf-16-le")
        start = pos


def grapheme_slice(text: str, start: int = 0, end: int = None) -> str:
    """
    Return a substring of text counting grapheme clusters not codepoints.
    """
    if start < 0 or (end is not None and end < 0):
        raise ValueError("negative start or end")

    s = 0
    slice_start: Optional[int] = None
    slice_end: Optional[int] = None
    for i, g in enumerate(graphemes(text)):
        if i == start:
            slice_start = s
        if i + 1 == end:
            slice_end = s + len(g)
            break
        s += len(g)
    if slice_start is None:
        return ""
    if slice_end is None:
        slice_end = len(text)
    return text[slice_start: slice_end]



import unittest

from orangewidget.tests.base import WidgetTest
from orangewidget.widget import OWBaseWidget, Msg


class TestMessages(WidgetTest):
    def test_clear_owner(self):
        class WidgetA(OWBaseWidget, openclass=True):
            class Error(OWBaseWidget.Error):
                err_a = Msg("error a")

        class WidgetB(WidgetA):
            class Error(WidgetA.Error):
                err_b = Msg("error b")

        w = self.create_widget(WidgetB)
        w.Error.err_a()
        w.Error.err_b()
        self.assertTrue(w.Error.err_a.is_shown())
        self.assertTrue(w.Error.err_b.is_shown())
        w.Error.clear()
        self.assertFalse(w.Error.err_a.is_shown())
        self.assertFalse(w.Error.err_b.is_shown())

        w.Error.err_a()
        w.Error.err_b()
        w.Error.clear(owner=WidgetB)
        self.assertTrue(w.Error.err_a.is_shown())
        self.assertFalse(w.Error.err_b.is_shown())

        w.Error.err_a()
        w.Error.err_b()
        w.Error.clear(owner=WidgetA)
        self.assertFalse(w.Error.err_a.is_shown())
        self.assertTrue(w.Error.err_b.is_shown())


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import date, datetime
from typing import Any, Dict

import numpy as np

from AnyQt.QtCore import Qt, QModelIndex, QLocale, QRect, QPoint, QSize
from AnyQt.QtGui import QStandardItemModel, QFont, QColor, QIcon, QImage, \
    QPainter
from AnyQt.QtWidgets import (
    QStyleOptionViewItem, QTableView, QAbstractItemDelegate
)

from orangecanvas.gui.svgiconengine import SvgIconEngine
from orangewidget.tests.base import GuiTest
from orangewidget.utils import graphemes, grapheme_slice
from orangewidget.utils.itemdelegates import ModelItemCache, \
    CachedDataItemDelegate, StyledItemDelegate, DataDelegate, \
    BarItemDataDelegate


def create_model(rows, columns):
    model = QStandardItemModel()
    model.setRowCount(rows)
    model.setColumnCount(columns)
    for i in range(rows):
        for j in range(columns):
            model.setItemData(
                model.index(i, j), {
                    Qt.DisplayRole: f"{i}x{j}",
                    Qt.UserRole: i * j,
                }
            )
    return model


class TestModelItemCache(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.model = create_model(10, 2)
        self.cache = ModelItemCache()

    def tearDown(self) -> None:
        del self.model
        del self.cache
        super().tearDown()

    def test_cache(self):
        model = self.model
        index = model.index(0, 0)
        res = self.cache.itemData(index, (Qt.DisplayRole, Qt.UserRole))
        self.assertEqual(res, {Qt.DisplayRole: "0x0", Qt.UserRole: 0})
        res = self.cache.itemData(index, (Qt.DisplayRole, Qt.UserRole,
                                          Qt.UserRole + 1))
        self.assertEqual(res, {Qt.DisplayRole: "0x0", Qt.UserRole: 0,
                               Qt.UserRole + 1: None})
        model.setData(index, "2", Qt.DisplayRole)
        res = self.cache.data(index, Qt.DisplayRole)
        self.assertEqual(res, "2")
        res = self.cache.data(index, Qt.UserRole + 2)
        self.assertIsNone(res)
        m1 = create_model(1, 1)
        res = self.cache.data(m1.index(0, 0), Qt.DisplayRole)
        self.assertEqual(res, "0x0")


class TestCachedDataItemDelegate(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.model = create_model(5, 2)
        self.delegate = CachedDataItemDelegate()

    def test_delegate(self):
        opt = QStyleOptionViewItem()
        index = self.model.index(0, 0)
        self.delegate.initStyleOption(opt, index)
        self.assertEqual(opt.text, "0x0")

        icon = QIcon(SvgIconEngine(b'<svg></svg>'))
        yellow = QColor(Qt.yellow)
        magenta = QColor(Qt.magenta)
        data = {
            Qt.DisplayRole: "AA",
            Qt.FontRole: QFont("Times New Roman"),
            Qt.TextAlignmentRole: Qt.AlignRight,
            Qt.CheckStateRole: Qt.Checked,
            Qt.DecorationRole: icon,
            Qt.ForegroundRole: yellow,
            Qt.BackgroundRole: magenta,
        }
        self.model.setItemData(index, data)
        self.delegate.initStyleOption(opt, index)
        self.assertEqual(opt.font.family(), QFont("Times New Roman").family())
        self.assertEqual(opt.displayAlignment, Qt.AlignRight)
        self.assertEqual(opt.backgroundBrush.color(), magenta)
        self.assertEqual(opt.palette.text().color(), yellow)
        self.assertFalse(opt.icon.isNull())
        self.assertEqual(opt.icon.cacheKey(), icon.cacheKey())

        res = self.delegate.cachedData(index, Qt.DisplayRole)
        self.assertEqual(res, "AA")
        res = self.delegate.cachedItemData(
            index, (Qt.DisplayRole, Qt.TextAlignmentRole)
        )
        self.assertIn(Qt.DisplayRole, res)
        self.assertIn(Qt.TextAlignmentRole, res)
        self.assertEqual(res[Qt.TextAlignmentRole], Qt.AlignRight)
        self.assertEqual(res[Qt.DisplayRole], "AA")


class TestStyledItemDelegate(unittest.TestCase):
    def test_display_text(self):
        delegate = StyledItemDelegate()
        locale = QLocale.c()
        displayText = lambda value: delegate.displayText(value, locale)
        self.assertEqual(displayText(None), "")
        self.assertEqual(displayText(1), "1")
        self.assertEqual(displayText(np.int64(1)), "1")
        self.assertEqual(displayText(np.int64(1)), "1")
        self.assertEqual(displayText(1.5), "1.5")
        self.assertEqual(displayText(np.float16(1.5)), "1.5")
        self.assertEqual(displayText("A"), "A")
        self.assertEqual(displayText(np.str_("A")), "A")

        self.assertEqual(displayText(date(1999, 12, 31)), "1999-12-31")
        self.assertEqual(displayText(datetime(1999, 12, 31, 23, 59, 59)),
                         "1999-12-31 23:59:59")

        self.assertEqual(displayText(np.datetime64(0, "s")),
                         "1970-01-01 00:00:00")


class TestDataDelegate(GuiTest):
    def setUp(self) -> None:
        super().setUp()
        self.view = QTableView()
        self.model = create_model(5, 2)
        self.delegate = DataDelegate(self.view)
        self.view.setItemDelegate(self.delegate)

    def tearDown(self) -> None:
        self.view.deleteLater()
        self.view = None
        self.model = None
        super().tearDown()

    def test_init_style_options(self):
        delegate = self.delegate
        model = self.model
        index = model.index(0, 0)
        model.setData(index, 1, Qt.DisplayRole)
        opt = QStyleOptionViewItem()
        delegate.initStyleOption(opt, index)
        self.assertEqual(opt.displayAlignment, Qt.AlignRight)
        model.setData(index, "A", Qt.DisplayRole)

        opt = QStyleOptionViewItem()
        delegate.initStyleOption(opt, index)
        self.assertEqual(opt.displayAlignment, Qt.AlignLeft)

    def test_paint(self):
        delegate = self.delegate
        model = self.model
        index = model.index(0, 0)
        model.setData(index, 1, Qt.DisplayRole)

        def paint_with_data(data):
            model.setItemData(index, data)
            opt = self.view.viewOptions()
            opt.rect = QRect(QPoint(0, 0), delegate.sizeHint(opt, index))
            delegate.initStyleOption(opt, index)
            img = QImage(opt.rect.size(), QImage.Format_ARGB32_Premultiplied)
            p = QPainter(img)
            try:
                delegate.paint(p, opt, index)
            finally:
                p.end()

        paint_with_data({Qt.DisplayRole: 1.0})
        paint_with_data({Qt.DisplayRole: "AA"})
        paint_with_data({Qt.DisplayRole: "AA",
                         Qt.TextAlignmentRole: Qt.AlignLeft | Qt.AlignTop})
        paint_with_data({Qt.DisplayRole: "AA",
                         Qt.TextAlignmentRole: Qt.AlignHCenter | Qt.AlignVCenter})
        paint_with_data({Qt.DisplayRole: "AA",
                         Qt.TextAlignmentRole: Qt.AlignRight | Qt.AlignBottom})

    def test_paint_long_combining(self):
        text = "ABC" * 1000
        opt = self.view.viewOptions()
        paint_with_data(self.delegate, {Qt.DisplayRole: text}, opt)
        text = "\N{TAMIL LETTER NA}\N{TAMIL VOWEL SIGN I}" * 10000
        paint_with_data(self.delegate, {Qt.DisplayRole: text}, opt)


class TestBarItemDataDelegate(GuiTest):
    def setUp(self) -> None:
        super().setUp()
        self.view = QTableView()
        self.model = create_model(5, 2)
        self.delegate = BarItemDataDelegate(self.view)
        self.view.setItemDelegate(self.delegate)

    def tearDown(self) -> None:
        self.view.deleteLater()
        self.view = None
        self.model = None
        super().tearDown()

    def test_size_hint(self):
        model = self.model
        index = model.index(0, 0)
        delegate = self.delegate
        model.setData(index, 0.5, delegate.barFillRatioRole)
        sh1 = delegate.sizeHint(self.view.viewOptions(), index)
        delegate.penWidth += 2
        sh2 = delegate.sizeHint(self.view.viewOptions(), index)
        self.assertGreater(sh2.height(), sh1.height())

    def test_paint(self):
        model = self.model
        index = model.index(0, 0)
        delegate = self.delegate
        model.setData(index, 0.5, delegate.barFillRatioRole)

        def paint_with_data_(data):
            paint_with_data(delegate, data, self.view.viewOptions())
            model.setItemData(index, data)
            opt = self.view.viewOptions()
            size = delegate.sizeHint(opt, index).expandedTo(QSize(10, 10))
            opt.rect = QRect(QPoint(0, 0), size)
            delegate.initStyleOption(opt, index)
            img = QImage(opt.rect.size(), QImage.Format_ARGB32_Premultiplied)
            p = QPainter(img)
            try:
                delegate.paint(p, opt, index)
            finally:
                p.end()
        paint_with_data_({delegate.barFillRatioRole: 0.2,
                          delegate.barColorRole: QColor(Qt.magenta)})
        paint_with_data_({delegate.barFillRatioRole: None,
                          delegate.barColorRole: None})


def paint_with_data(
        delegate: QAbstractItemDelegate,
        data: Dict[int, Any],
        options: QStyleOptionViewItem = None
) -> None:
    model = create_model(1, 1)
    index = model.index(0, 0)
    model.setItemData(index, data)
    opt = QStyleOptionViewItem(options) if options is not None else QStyleOptionViewItem()
    size = delegate.sizeHint(opt, index).expandedTo(QSize(10, 10))
    opt.rect = QRect(QPoint(0, 0), size)
    delegate.initStyleOption(opt, index)
    img = QImage(opt.rect.size(), QImage.Format_ARGB32_Premultiplied)
    p = QPainter(img)
    try:
        delegate.paint(p, opt, index)
    finally:
        p.end()


class TestGraphemes(unittest.TestCase):
    def test_grapheme(self):
        self.assertEqual(list(graphemes("")), [])
        self.assertEqual(list(graphemes("a")), ["a"])
        self.assertEqual(list(graphemes("ab")), ["a", "b"])
        text = "\N{TAMIL LETTER NA}\N{TAMIL VOWEL SIGN I}"
        self.assertEqual(list(graphemes(text)), [text])
        self.assertEqual(list(graphemes("a" + text)), ["a", text])
        self.assertEqual(list(graphemes(text + "b")), [text, "b"])
        self.assertEqual(list(graphemes("a" + text + "b")), ["a", text, "b"])
        self.assertEqual(list(graphemes("a" + text + "b" + text)), ["a", text, "b", text])
        self.assertEqual(list(graphemes("a" + text + text + "b")), ["a", text, text, "b"])

    def test_grapheme_slice(self):
        self.assertEqual(grapheme_slice(""), "")
        self.assertEqual(grapheme_slice("", start=1), "")
        self.assertEqual(grapheme_slice("", end=1), "")
        self.assertEqual(grapheme_slice("a"), "a")
        self.assertEqual(grapheme_slice("a", start=1), "")
        self.assertEqual(grapheme_slice("ab"), "ab")
        self.assertEqual(grapheme_slice("ab", start=1), "b")
        self.assertEqual(grapheme_slice("ab", end=1), "a")
        self.assertEqual(grapheme_slice("ab", start=1, end=1), "")
        self.assertEqual(grapheme_slice("abc", start=1, end=2), "b")
        text = "\N{TAMIL LETTER NA}\N{TAMIL VOWEL SIGN I}"
        self.assertEqual(grapheme_slice(text * 3, end=2), text * 2)
        self.assertEqual(grapheme_slice("a" + text + "b" + text, end=3),
                         "a" + text + "b")



"""
General helper functions and classes for PyQt concurrent programming
"""
# TODO: Rename the module to something that does not conflict with stdlib
# concurrent
from typing import Callable, Any, List, Optional
import threading
import logging
import warnings
import weakref
from functools import partial
import concurrent.futures
from concurrent.futures import Future, TimeoutError

from AnyQt.QtCore import (
    Qt, QObject, QMetaObject, QThreadPool, QThread, QRunnable, QSemaphore,
    QCoreApplication, QEvent, Q_ARG,
    pyqtSignal as Signal, pyqtSlot as Slot
)
from AnyQt import sip

_log = logging.getLogger(__name__)


class PyOwned:
    """
    A mixin for python owned QObject's used as queued cross thread
    communication channels.

    When this object is released from a thread that is not self.thread()
    it is *resurrected* and scheduled for deferred deletion from its own
    thread with self.deleteLater()
    """
    # This is a workaround for:
    # https://www.riverbankcomputing.com/pipermail/pyqt/2020-April/042734.html
    # Should not be necessary with PyQt5-sip>=12.8 (i.e sip api 12.8)
    __delete_later_set = set()

    def __del__(self: QObject):
        # Note: This is otherwise quite similar to how PyQt5 does this except
        # for the resurrection (i.e. the wrapper is allowed to be freed, but
        # C++ part is deleteLater-ed).
        if sip.ispyowned(self):
            try:
                own_thread = self.thread() is QThread.currentThread()
            except RuntimeError:
                return
            if not own_thread:
                # object resurrection; keep python wrapper alive and schedule
                # deletion from the object's own thread.
                PyOwned.__delete_later_set.add(self)
                ref = weakref.ref(self)

                # Clear final ref from 'destroyed' signal. As late as possible
                # in QObject' destruction.
                def clear():
                    self = ref()
                    try:
                        PyOwned.__delete_later_set.remove(self)
                    except KeyError:
                        pass
                self.destroyed.connect(clear, Qt.DirectConnection)
                self.deleteLater()


class FutureRunnable(QRunnable):
    """
    A QRunnable to fulfil a `Future` in a QThreadPool managed thread.

    Parameters
    ----------
    future : concurrent.futures.Future
        Future whose contents will be set with the result of executing
        `func(*args, **kwargs)` after completion
    func : Callable
        Function to invoke in a thread
    args : tuple
        Positional arguments for `func`
    kwargs : dict
        Keyword arguments for `func`

    Example
    -------
    >>> f = concurrent.futures.Future()
    >>> task = FutureRunnable(f, int, (42,), {})
    >>> QThreadPool.globalInstance().start(task)
    >>> f.result()
    42
    """
    def __init__(self, future, func, args, kwargs):
        # type: (Future, Callable, tuple, dict) -> None
        super().__init__()
        self.future = future
        self.task = (func, args, kwargs)

    def run(self):
        """
        Reimplemented from `QRunnable.run`
        """
        try:
            if not self.future.set_running_or_notify_cancel():
                # future was cancelled
                return
            func, args, kwargs = self.task
            try:
                result = func(*args, **kwargs)
            except BaseException as ex: # pylint: disable=broad-except
                self.future.set_exception(ex)
            else:
                self.future.set_result(result)
        except BaseException:  # pylint: disable=broad-except
            log = logging.getLogger(__name__)
            log.critical("Exception in worker thread.", exc_info=True)


class FutureWatcher(QObject, PyOwned):
    """
    An `QObject` watching the state changes of a `concurrent.futures.Future`

    Note
    ----
    The state change notification signals (`done`, `finished`, ...)
    are always emitted when the control flow reaches the event loop
    (even if the future is already completed when set).

    Note
    ----
    An event loop must be running, otherwise the notifier signals will
    not be emitted.

    Parameters
    ----------
    parent : QObject
        Parent object.
    future : Future
        The future instance to watch.

    Example
    -------
    >>> app = QCoreApplication.instance() or QCoreApplication([])
    >>> f = submit(lambda i, j: i ** j, 10, 3)
    >>> watcher = FutureWatcher(f)
    >>> watcher.resultReady.connect(lambda res: print("Result:", res))
    >>> watcher.done.connect(app.quit)
    >>> _ = app.exec()
    Result: 1000
    >>> f.result()
    1000
    """
    #: Signal emitted when the future is done (cancelled or finished)
    done = Signal(Future)

    #: Signal emitted when the future is finished (i.e. returned a result
    #: or raised an exception - but not if cancelled)
    finished = Signal(Future)

    #: Signal emitted when the future was cancelled
    cancelled = Signal(Future)

    #: Signal emitted with the future's result when successfully finished.
    resultReady = Signal(object)

    #: Signal emitted with the future's exception when finished with an
    #: exception.
    exceptionReady = Signal(BaseException)

    # A private event type used to notify the watcher of a Future's completion
    __FutureDone = QEvent.Type(QEvent.registerEventType())

    def __init__(self, future=None, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.__future = None
        if future is not None:
            self.setFuture(future)

    def setFuture(self, future):
        # type: (Future) -> None
        """
        Set the future to watch.

        Raise a `RuntimeError` if a future is already set.

        Parameters
        ----------
        future : Future
        """
        if self.__future is not None:
            raise RuntimeError("Future already set")

        self.__future = future
        selfweakref = weakref.ref(self)

        def on_done(f):
            assert f is future
            selfref = selfweakref()

            if selfref is None:
                return

            try:
                QCoreApplication.postEvent(
                    selfref, QEvent(FutureWatcher.__FutureDone))
            except RuntimeError:
                # Ignore RuntimeErrors (when C++ side of QObject is deleted)
                # (? Use QObject.destroyed and remove the done callback ?)
                pass

        future.add_done_callback(on_done)

    def future(self):
        # type: () -> Future
        """
        Return the future instance.
        """
        return self.__future

    def isCancelled(self):
        warnings.warn("isCancelled is deprecated", DeprecationWarning,
                      stacklevel=2)
        return self.__future.cancelled()

    def isDone(self):
        warnings.warn("isDone is deprecated", DeprecationWarning,
                      stacklevel=2)
        return self.__future.done()

    def result(self):
        # type: () -> Any
        """
        Return the future's result.

        Note
        ----
        This method is non-blocking. If the future has not yet completed
        it will raise an error.
        """
        try:
            return self.__future.result(timeout=0)
        except TimeoutError:
            raise RuntimeError("Future is not yet done")

    def exception(self):
        # type: () -> Optional[BaseException]
        """
        Return the future's exception.

        Note
        ----
        This method is non-blocking. If the future has not yet completed
        it will raise an error.
        """
        try:
            return self.__future.exception(timeout=0)
        except TimeoutError:
            raise RuntimeError("Future is not yet done")

    def __emitSignals(self):
        assert self.__future is not None
        assert self.__future.done()
        if self.__future.cancelled():
            self.cancelled.emit(self.__future)
            self.done.emit(self.__future)
        elif self.__future.done():
            self.finished.emit(self.__future)
            self.done.emit(self.__future)
            if self.__future.exception():
                self.exceptionReady.emit(self.__future.exception())
            else:
                self.resultReady.emit(self.__future.result())
        else:
            assert False

    def customEvent(self, event):
        # Reimplemented.
        if event.type() == FutureWatcher.__FutureDone:
            self.__emitSignals()
        super().customEvent(event)


class FutureSetWatcher(QObject, PyOwned):
    """
    An `QObject` watching the state changes of a list of
    `concurrent.futures.Future` instances

    Note
    ----
    The state change notification signals (`doneAt`, `finishedAt`, ...)
    are always emitted when the control flow reaches the event loop
    (even if the future is already completed when set).

    Note
    ----
    An event loop must be running, otherwise the notifier signals will
    not be emitted.

    Parameters
    ----------
    parent : QObject
        Parent object.
    futures : List[Future]
        A list of future instance to watch.

    Example
    -------
    >>> app = QCoreApplication.instance() or QCoreApplication([])
    >>> fs = [submit(lambda i, j: i ** j, 10, 3) for i in range(10)]
    >>> watcher = FutureSetWatcher(fs)
    >>> watcher.resultReadyAt.connect(
    ...     lambda i, res: print("Result at {}: {}".format(i, res))
    ... )
    >>> watcher.doneAll.connect(app.quit)
    >>> _ = app.exec()
    Result at 0: 1000
    ...
    """
    #: Signal emitted when the future at `index` is done (cancelled or
    #: finished)
    doneAt = Signal([int, Future])

    #: Signal emitted when the future at index is finished (i.e. returned
    #: a result)
    finishedAt = Signal([int, Future])

    #: Signal emitted when the future at `index` was cancelled.
    cancelledAt = Signal([int, Future])

    #: Signal emitted with the future's result when successfully
    #: finished.
    resultReadyAt = Signal([int, object])

    #: Signal emitted with the future's exception when finished with an
    #: exception.
    exceptionReadyAt = Signal([int, BaseException])

    #: Signal reporting the current completed count
    progressChanged = Signal([int, int])

    #: Signal emitted when all the futures have completed.
    doneAll = Signal()

    def __init__(self, futures: Optional[List['Future']] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__futures = None
        self.__semaphore = None
        self.__countdone = 0
        if futures is not None:
            self.setFutures(futures)

    def setFutures(self, futures):
        # type: (List[Future]) -> None
        """
        Set the future instances to watch.

        Raise a `RuntimeError` if futures are already set.

        Parameters
        ----------
        futures : List[Future]
        """
        if self.__futures is not None:
            raise RuntimeError("already set")
        self.__futures = []
        selfweakref = weakref.ref(self)
        schedule_emit = methodinvoke(self, "__emitpending", (int, Future))

        # Semaphore counting the number of future that have enqueued
        # done notifications. Used for the `wait` implementation.
        self.__semaphore = semaphore = QSemaphore(0)

        for i, future in enumerate(futures):
            self.__futures.append(future)

            def on_done(index, f):
                try:
                    selfref = selfweakref()  # not safe really
                    if selfref is None:  # pragma: no cover
                        return
                    try:
                        schedule_emit(index, f)
                    except RuntimeError:  # pragma: no cover
                        # Ignore RuntimeErrors (when C++ side of QObject is deleted)
                        # (? Use QObject.destroyed and remove the done callback ?)
                        pass
                finally:
                    semaphore.release()

            future.add_done_callback(partial(on_done, i))

        if not self.__futures:
            # `futures` was an empty sequence.
            methodinvoke(self, "doneAll", ())()

    @Slot(int, Future)
    def __emitpending(self, index, future):
        # type: (int, Future) -> None
        assert QThread.currentThread() is self.thread()
        assert self.__futures[index] is future
        assert future.done()
        assert self.__countdone < len(self.__futures)
        self.__futures[index] = None
        self.__countdone += 1

        if future.cancelled():
            self.cancelledAt.emit(index, future)
            self.doneAt.emit(index, future)
        elif future.done():
            self.finishedAt.emit(index, future)
            self.doneAt.emit(index, future)
            if future.exception():
                self.exceptionReadyAt.emit(index, future.exception())
            else:
                self.resultReadyAt.emit(index, future.result())
        else:
            assert False

        self.progressChanged.emit(self.__countdone, len(self.__futures))

        if self.__countdone == len(self.__futures):
            self.doneAll.emit()

    def flush(self):
        """
        Flush all pending signal emits currently enqueued.

        Must only ever be called from the thread this object lives in
        (:func:`QObject.thread()`).
        """
        if QThread.currentThread() is not self.thread():
            raise RuntimeError("`flush()` called from a wrong thread.")
        # NOTE: QEvent.MetaCall is the event implementing the
        # `Qt.QueuedConnection` method invocation.
        QCoreApplication.sendPostedEvents(self, QEvent.MetaCall)

    def wait(self):
        """
        Wait for for all the futures to complete and *enqueue* notifications
        to this object, but do not emit any signals.

        Use `flush()` to emit all signals after a `wait()`
        """
        if self.__futures is None:
            raise RuntimeError("Futures were not set.")

        self.__semaphore.acquire(len(self.__futures))
        self.__semaphore.release(len(self.__futures))


class methodinvoke(object):
    """
    A thin wrapper for invoking QObject's method through
    `QMetaObject.invokeMethod`.

    This can be used to invoke the method across thread boundaries (or even
    just for scheduling delayed calls within the same thread).

    Note
    ----
    An event loop MUST be running in the target QObject's thread.

    Parameters
    ----------
    obj : QObject
        A QObject instance.
    method : str
        The method name. This method must be registered with the Qt object
        meta system (e.g. decorated by a Slot decorator).
    arg_types : tuple
        A tuple of positional argument types.
    conntype : Qt.ConnectionType
        The connection/call type. Qt.QueuedConnection (the default) and
        Qt.BlockingConnection are the most interesting.

    See Also
    --------
    QMetaObject.invokeMethod

    Example
    -------
    >>> app = QCoreApplication.instance() or QCoreApplication([])
    >>> quit = methodinvoke(app, "quit", ())
    >>> t = threading.Thread(target=quit)
    >>> t.start()
    >>> app.exec()
    0
    """
    @staticmethod
    def from_method(method, arg_types=(), *, conntype=Qt.QueuedConnection):
        """
        Create and return a `methodinvoke` instance from a bound method.

        Parameters
        ----------
        method : Union[types.MethodType, types.BuiltinMethodType]
            A bound method of a QObject registered with the Qt meta object
            system (e.g. decorated by a Slot decorators)
        arg_types : Tuple[Union[type, str]]
            A tuple of positional argument types.
        conntype: Qt.ConnectionType
            The connection/call type (Qt.QueuedConnection and
            Qt.BlockingConnection are the most interesting)

        Returns
        -------
        invoker : methodinvoke
        """
        obj = method.__self__
        name = method.__name__
        return methodinvoke(obj, name, arg_types, conntype=conntype)

    def __init__(self, obj, method, arg_types=(), *,
                 conntype=Qt.QueuedConnection):
        self.obj = obj
        self.method = method
        self.arg_types = tuple(arg_types)
        self.conntype = conntype

    def __call__(self, *args):
        args = [Q_ARG(atype, arg) for atype, arg in zip(self.arg_types, args)]
        return QMetaObject.invokeMethod(
            self.obj, self.method, self.conntype, *args)



import collections
from collections import defaultdict
from typing import Sequence
from math import isnan, isinf
from numbers import Number, Integral

import operator

from contextlib import contextmanager
from warnings import warn

from AnyQt.QtCore import (
    Qt, QObject, QAbstractListModel, QAbstractTableModel, QModelIndex,
    QItemSelectionModel, QMimeData
)
from AnyQt.QtCore import pyqtSignal as Signal
from AnyQt.QtWidgets import (
    QWidget, QBoxLayout, QToolButton, QAbstractButton, QAction,
    QStyledItemDelegate
)
from AnyQt.QtGui import QPalette, QPen

import numpy


class _store(dict):
    pass


def _argsort(seq, cmp=None, key=None, reverse=False):
    indices = range(len(seq))
    if key is not None:
        return sorted(indices, key=lambda i: key(seq[i]), reverse=reverse)
    elif cmp is not None:
        from functools import cmp_to_key
        return sorted(indices, key=cmp_to_key(lambda a, b: cmp(seq[a], seq[b])),
                      reverse=reverse)
    else:
        return sorted(indices, key=lambda i: seq[i], reverse=reverse)


@contextmanager
def signal_blocking(obj):
    blocked = obj.signalsBlocked()
    obj.blockSignals(True)
    try:
        yield
    finally:
        obj.blockSignals(blocked)


def _as_contiguous_range(the_slice, length):
    start, stop, step = the_slice.indices(length)
    if step == -1:
        # Equivalent range with positive step
        start, stop, step = stop + 1, start + 1, 1
    elif not (step == 1 or step is None):
        raise IndexError("Non-contiguous range.")
    return start, stop, step


class AbstractSortTableModel(QAbstractTableModel):
    """
    A sorting proxy table model that sorts its rows in fast numpy,
    avoiding potentially thousands of calls into
    ``QSortFilterProxyModel.lessThan()`` or any potentially costly
    reordering of original data.

    Override ``sortColumnData()``, adapting it to your underlying model.

    Make sure to use ``mapToSourceRows()``/``mapFromSourceRows()``
    whenever fetching or manipulating table data, such as in ``data()``.

    When updating the model (inserting, removing rows), the sort order
    needs to be accounted for (e.g. reset and re-applied).
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__sortInd = None     #: Indices sorting the source table
        self.__sortIndInv = None  #: The inverse of __sortInd
        self.__sortColumn = -1    #: Sort key column, or -1
        self.__sortOrder = Qt.AscendingOrder

    def sortColumnData(self, column):
        """Return raw, sortable data for column"""
        raise NotImplementedError

    def _sortColumnData(self, column):
        try:
            # Call the overridden implementation if available
            data = numpy.asarray(self.sortColumnData(column))
            data = data[self.mapToSourceRows(Ellipsis)]
        except NotImplementedError:
            # Fallback to slow implementation
            data = numpy.array([self.index(row, column).data()
                                for row in range(self.rowCount())])
        assert data.ndim in (1, 2), 'Data should be 1- or 2-dimensional'
        return data

    def sortColumn(self):
        """The column currently used for sorting (-1 if no sorting is applied)"""
        return self.__sortColumn

    def sortOrder(self):
        """The current sort order"""
        return self.__sortOrder

    def mapToSourceRows(self, rows):
        """Return array of row indices in the source table for given model rows

        Parameters
        ----------
        rows : int or list of int or numpy.ndarray of dtype=int or Ellipsis
            View (sorted) rows.

        Returns
        -------
        numpy.ndarray
            Source rows matching input rows. If they are the same,
            simply input `rows` is returned.
        """
        # self.__sortInd[rows] fails if `rows` is an empty list or array
        if self.__sortInd is not None \
                and (isinstance(rows, (Integral, type(Ellipsis)))
                     or len(rows)):
            new_rows = self.__sortInd[rows]
            if rows is Ellipsis:
                new_rows.setflags(write=False)
            rows = new_rows
        return rows

    def mapFromSourceRows(self, rows):
        """Return array of row indices in the model for given source table rows

        Parameters
        ----------
        rows : int or list of int or numpy.ndarray of dtype=int or Ellipsis
            Source model rows.

        Returns
        -------
        numpy.ndarray
            ModelIndex (sorted) rows matching input source rows.
            If they are the same, simply input `rows` is returned.
        """
        # self.__sortInd[rows] fails if `rows` is an empty list or array
        if self.__sortIndInv is not None \
                and (isinstance(rows, (Integral, type(Ellipsis)))
                     or len(rows)):
            new_rows = self.__sortIndInv[rows]
            if rows is Ellipsis:
                new_rows.setflags(write=False)
            rows = new_rows
        return rows

    def resetSorting(self):
        """Invalidates the current sorting"""
        return self.sort(-1)

    def _argsortData(self, data: numpy.ndarray, order):
        """
        Return indices of sorted data. May be reimplemented to handle
        sorting in a certain way, e.g. to sort NaN values last.
        """
        if order == Qt.DescendingOrder:
            # to ensure stable descending order, sort reversed data ...
            data = data[::-1]
        if data.ndim == 1:
            indices = numpy.argsort(data, kind="mergesort")
        else:
            indices = numpy.lexsort(data.T[::-1])
        if order == Qt.DescendingOrder:
            # ... and reverse (as well as invert) resulting indices
            indices = len(data) - 1 - indices[::-1]
        return indices

    def sort(self, column: int, order: Qt.SortOrder = Qt.AscendingOrder):
        """
        Sort the data by `column` into `order`.

        To reset the order, pass column=-1.

        Reimplemented from QAbstractItemModel.sort().

        Notes
        -----
        This only affects the model's data presentation. The underlying
        data table is left unmodified. Use mapToSourceRows()/mapFromSourceRows()
        when accessing data by row indexes.
        """
        indices = self._sort(column, order)
        self.__sortColumn = -1 if column < 0 else column
        self.__sortOrder = order
        self.setSortIndices(indices)

    def setSortIndices(self, indices):
        self.layoutAboutToBeChanged.emit([], QAbstractTableModel.VerticalSortHint)

        # Store persistent indices as well as their (actual) rows in the
        # source data table.
        persistent = self.persistentIndexList()
        persistent_rows = self.mapToSourceRows([i.row() for i in persistent])

        if indices is not None:
            self.__sortInd = numpy.asarray(indices)
            self.__sortIndInv = numpy.argsort(indices)
        else:
            self.__sortInd = None
            self.__sortIndInv = None

        persistent_rows = self.mapFromSourceRows(persistent_rows)

        self.changePersistentIndexList(
            persistent,
            [self.index(row, pind.column())
             for row, pind in zip(persistent_rows, persistent)])
        self.layoutChanged.emit([], QAbstractTableModel.VerticalSortHint)

    def _sort(self, column, order):
        indices = None
        if column >= 0:
            # - _sortColumnData returns data in its currently shown order
            # - _argSortData thus returns an array a, in which a[i] is the row
            #   number (in the current view) to put to line i
            # - mapToSourceRows maps these indices back to original data.
            # This contrived procedure (instead of _sortColumnData returning
            # the original data, saving us from double mapping) ensures stable
            # sort order on consecutive calls
            data = numpy.asarray(self._sortColumnData(column))
            if data is None:
                data = numpy.arange(self.rowCount())
            elif data.dtype == object:
                data = data.astype(str)
            indices = self.mapToSourceRows(self._argsortData(data, order))
        return indices


class PyTableModel(AbstractSortTableModel):
    """ A model for displaying python tables (sequences of sequences) in
    QTableView objects.

    Parameters
    ----------
    sequence : list
        The initial list to wrap.
    parent : QObject
        Parent QObject.
    editable: bool or sequence
        If True, all items are flagged editable. If sequence, the True-ish
        fields mark their respective columns editable.

    Notes
    -----
    The model rounds numbers to human readable precision, e.g.:
    1.23e-04, 1.234, 1234.5, 12345, 1.234e06.

    To set additional item roles, use setData().
    """

    @staticmethod
    def _RoleData():
        return defaultdict(lambda: defaultdict(dict))

    def __init__(self, sequence=None, parent=None, editable=False):
        super().__init__(parent)
        self._headers = {}
        self._editable = editable
        self._table = None
        self._roleData = None
        if sequence is None:
            sequence = []
        self.wrap(sequence)

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self)

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else max(map(len, self._table), default=0)

    def flags(self, index):
        flags = super().flags(index)
        if not self._editable or not index.isValid():
            return flags
        if isinstance(self._editable, Sequence):
            return flags | Qt.ItemIsEditable if self._editable[index.column()] else flags
        return flags | Qt.ItemIsEditable

    def setData(self, index, value, role=Qt.EditRole):
        row = self.mapFromSourceRows(index.row())
        if role == Qt.EditRole:
            self[row][index.column()] = value
            self.dataChanged.emit(index, index)
        else:
            self._roleData[row][index.column()][role] = value
        return True

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid():
            return

        row, column = self.mapToSourceRows(index.row()), index.column()

        role_value = self._roleData.get(row, {}).get(column, {}).get(role)
        if role_value is not None:
            return role_value

        try:
            value = self[row][column]
        except IndexError:
            return
        if role == Qt.EditRole:
            return value
        # if role == Qt.DecorationRole and isinstance(value, Variable):
        #     return gui.attributeIconDict[value]
        if role == Qt.DisplayRole:
            if (isinstance(value, Number) and
                    not (isnan(value) or isinf(value) or isinstance(value, Integral))):
                absval = abs(value)
                strlen = len(str(int(absval)))
                value = '{:.{}{}}'.format(value,
                                          2 if absval < .001 else
                                          3 if strlen < 2 else
                                          1 if strlen < 5 else
                                          0 if strlen < 6 else
                                          3,
                                          'f' if (absval == 0 or
                                                  absval >= .001 and
                                                  strlen < 6)
                                          else 'e')
            return str(value)
        if role == Qt.TextAlignmentRole and isinstance(value, Number):
            return Qt.AlignRight | Qt.AlignVCenter
        if role == Qt.ToolTipRole:
            return str(value)

    def sortColumnData(self, column):
        return [row[column] for row in self._table]

    def setHorizontalHeaderLabels(self, labels):
        """
        Parameters
        ----------
        labels : list of str
        """
        self._headers[Qt.Horizontal] = tuple(labels)

    def setVerticalHeaderLabels(self, labels):
        """
        Parameters
        ----------
        labels : list of str
        """
        self._headers[Qt.Vertical] = tuple(labels)

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        headers = self._headers.get(orientation)

        if headers and section < len(headers):
            section = self.mapToSourceRows(section) if orientation == Qt.Vertical else section
            value = headers[section]
            if role == Qt.ToolTipRole:
                role = Qt.DisplayRole
            if role == Qt.DisplayRole:
                return value

        # Use QAbstractItemModel default for non-existent header/sections
        return super().headerData(section, orientation, role)

    def removeRows(self, row, count, parent=QModelIndex()):
        if not parent.isValid():
            del self[row:row + count]
            for rowidx in range(row, row + count):
                self._roleData.pop(rowidx, None)
            return True
        return False

    def removeColumns(self, column, count, parent=QModelIndex()):
        self.beginRemoveColumns(parent, column, column + count - 1)
        for row in self._table:
            del row[column:column + count]
        for cols in self._roleData.values():
            for col in range(column, column + count):
                cols.pop(col, None)
        del self._headers.get(Qt.Horizontal, [])[column:column + count]
        self.endRemoveColumns()
        return True

    def insertRows(self, row, count, parent=QModelIndex()):
        self.beginInsertRows(parent, row, row + count - 1)
        self._table[row:row] = [[''] * self.columnCount() for _ in range(count)]
        self.endInsertRows()
        return True

    def insertColumns(self, column, count, parent=QModelIndex()):
        self.beginInsertColumns(parent, column, column + count - 1)
        for row in self._table:
            row[column:column] = [''] * count
        self.endInsertColumns()
        return True

    def __len__(self):
        return len(self._table)

    def __bool__(self):
        return len(self) != 0

    def __iter__(self):
        return iter(self._table)

    def __getitem__(self, item):
        return self._table[item]

    def __delitem__(self, i):
        if isinstance(i, slice):
            start, stop, _ = _as_contiguous_range(i, len(self))
            stop -= 1
        else:
            start = stop = i = i if i >= 0 else len(self) + i
        self._check_sort_order()
        self.beginRemoveRows(QModelIndex(), start, stop)
        del self._table[i]
        self.endRemoveRows()

    def __setitem__(self, i, value):
        self._check_sort_order()

        if isinstance(i, slice):
            start, stop, _ = _as_contiguous_range(i, len(self))
            if not isinstance(value, collections.abc.Sized):
                value = tuple(value)
            newstop = start + len(value)

            # Signal changes
            parent = QModelIndex()
            if newstop > stop:
                self.rowsAboutToBeInserted.emit(parent, stop, newstop - 1)
            elif newstop < stop:
                self.rowsAboutToBeRemoved.emit(parent, newstop, stop - 1)

            # Make changes
            self._table[i] = value

            # Signal change were made
            if start != min(stop, newstop):
                self.dataChanged.emit(
                    self.index(start, 0),
                    self.index(min(stop, newstop) - 1, self.columnCount() - 1))
            if newstop > stop:
                self.rowsInserted.emit(parent, stop, newstop - 1)
            elif newstop < stop:
                self.rowsRemoved.emit(parent, newstop, stop - 1)
        else:
            self._table[i] = value
            i %= len(self)
            self.dataChanged.emit(self.index(i, 0),
                                  self.index(i, self.columnCount() - 1))

    def _check_sort_order(self):
        if self.mapToSourceRows(Ellipsis) is not Ellipsis:
            warn("Can't modify PyTableModel when it's sorted",
                 RuntimeWarning, stacklevel=3)
            raise RuntimeError("Can't modify PyTableModel when it's sorted")

    def wrap(self, table):
        self.beginResetModel()
        self._table = table
        self._roleData = self._RoleData()
        self.resetSorting()
        self.endResetModel()

    def tolist(self):
        return self._table

    def clear(self):
        self.beginResetModel()
        self._table.clear()
        self.resetSorting()
        self._roleData.clear()
        self.endResetModel()

    def append(self, row):
        self.extend([row])

    def _insertColumns(self, rows):
        n_max = max(map(len, rows))
        if self.columnCount() < n_max:
            self.insertColumns(self.columnCount(), n_max - self.columnCount())

    def extend(self, rows):
        i, rows = len(self), list(rows)
        self.insertRows(i, len(rows))
        self._insertColumns(rows)
        self[i:] = rows

    def insert(self, i, row):
        self.insertRows(i, 1)
        self._insertColumns((row,))
        self[i] = row

    def remove(self, val):
        del self[self._table.index(val)]


class SeparatorItem:
    pass


class LabelledSeparator(SeparatorItem):
    def __init__(self, label=None):
        self.label = label


class SeparatedListDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        # type: (QPainter, QStyleOptionViewItem, QModelIndex) -> None
        super().paint(painter, option, index)
        data = index.data(Qt.EditRole)
        if not isinstance(data, LabelledSeparator):
            return

        painter.save()
        palette = option.palette  # type: QPalette
        rect = option.rect  # type: QRect
        if data.label:
            y = int(rect.bottom() - 0.1 * rect.height())
            brush = palette.brush(QPalette.Active, QPalette.WindowText)
            font = painter.font()
            font.setPointSizeF(font.pointSizeF() * 0.9)
            font.setBold(True)
            painter.setFont(font)
            painter.setPen(QPen(brush, 1.0))
            painter.drawText(rect, Qt.AlignCenter, data.label)
        else:
            y = rect.center().y()
        brush = palette.brush(QPalette.Disabled, QPalette.WindowText)
        painter.setPen(QPen(brush, 1.0))
        painter.drawLine(rect.left(), y, rect.left() + rect.width(), y)
        painter.restore()


class PyListModel(QAbstractListModel):
    """ A model for displaying python list like objects in Qt item view classes
    """
    MIME_TYPE = "application/x-Orange-PyListModelData"
    Separator = SeparatorItem()
    removed = Signal()

    def __init__(self, iterable=None, parent=None,
                 flags=Qt.ItemIsSelectable | Qt.ItemIsEnabled,
                 list_item_role=Qt.DisplayRole,
                 enable_dnd=False,
                 supportedDropActions=Qt.MoveAction):
        super().__init__(parent)
        self._list = []
        self._other_data = []
        if enable_dnd:
            flags |= Qt.ItemIsDragEnabled
        self._flags = flags
        self.list_item_role = list_item_role

        self._supportedDropActions = supportedDropActions
        if iterable is not None:
            self.extend(iterable)

    def _is_index_valid(self, index):
        # This error would happen if one wraps a list into a model and then
        # modifies a list instead of a model
        if len(self) != len(self._other_data):
            raise RuntimeError("Mismatched length of model and its _other_data")
        if isinstance(index, QModelIndex) and index.isValid():
            row, column = index.row(), index.column()
            return 0 <= row < len(self) and column == 0
        elif isinstance(index, int):
            return -len(self) <= index < len(self)
        else:
            return False

    def wrap(self, lst):
        """ Wrap the list with this model. All changes to the model
        are done in place on the passed list
        """
        self.beginResetModel()
        self._list = lst
        self._other_data = [_store() for _ in lst]
        self.endResetModel()

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            return str(section)

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self._list)

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else 1

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if role in [self.list_item_role, Qt.EditRole] \
                and self._is_index_valid(index):
            return self[row]
        elif self._is_index_valid(row):
            if isinstance(self[row], SeparatorItem) \
                    and role == Qt.AccessibleDescriptionRole:
                return 'separator'
            return self._other_data[row].get(role, None)

    def itemData(self, index):
        mapping = QAbstractListModel.itemData(self, index)
        if self._is_index_valid(index):
            items = list(self._other_data[index.row()].items())
        else:
            items = []
        for key, value in items:
            mapping[key] = value
        return mapping

    def parent(self, index=QModelIndex()):
        return QModelIndex()

    def setData(self, index, value, role=Qt.EditRole):
        if role == Qt.EditRole:
            if self._is_index_valid(index):
                self._list[index.row()] = value
                self.dataChanged.emit(index, index)
                return True
        elif self._is_index_valid(index):
            self._other_data[index.row()][role] = value
            self.dataChanged.emit(index, index)
            return True
        return False

    def setItemData(self, index, data):
        data = dict(data)
        if not data:
            return True  # pragma: no cover

        with signal_blocking(self):
            for role, value in data.items():
                if role == Qt.EditRole and \
                        self._is_index_valid(index):
                    self._list[index.row()] = value
                elif self._is_index_valid(index):
                    self._other_data[index.row()][role] = value

        self.dataChanged.emit(index, index)
        return True

    def flags(self, index):
        if self._is_index_valid(index):
            row = index.row()
            default = Qt.NoItemFlags \
                if isinstance(self[row], SeparatorItem) else self._flags
            return self._other_data[row].get("flags", default)
        else:
            return self._flags | Qt.ItemIsDropEnabled

    def insertRows(self, row, count, parent=QModelIndex()):
        """ Insert ``count`` rows at ``row``, the list fill be filled
        with ``None``
        """
        if not parent.isValid():
            self[row:row] = [None] * count
            return True
        else:
            return False

    def removeRows(self, row, count, parent=QModelIndex()):
        """Remove ``count`` rows starting at ``row``
        """
        if not parent.isValid():
            del self[row:row + count]
            self.removed.emit()
            return True
        else:
            return False

    def moveRows(self, sourceParent, sourceRow, count,
                 destinationParent, destinationChild):
        # type: (QModelIndex, int, int, QModelIndex, int) -> bool
        """
        Move `count` rows starting at `sourceRow` to `destinationChild`.

        Reimplemented from QAbstractItemModel.moveRows
        """
        if not self.beginMoveRows(sourceParent, sourceRow, sourceRow + count - 1,
                                  destinationParent, destinationChild):
            return False
        take_slice = slice(sourceRow, sourceRow + count)
        insert_at = destinationChild
        if insert_at > sourceRow:
            insert_at -= count
        items, other = self._list[take_slice], self._other_data[take_slice]
        del self._list[take_slice], self._other_data[take_slice]
        self._list[insert_at:insert_at] = items
        self._other_data[insert_at: insert_at] = other
        self.endMoveRows()
        return True

    def extend(self, iterable):
        list_ = list(iterable)
        count = len(list_)
        if count == 0:
            return
        self.beginInsertRows(QModelIndex(),
                             len(self), len(self) + count - 1)
        self._list.extend(list_)
        self._other_data.extend([_store() for _ in list_])
        self.endInsertRows()

    def append(self, item):
        self.extend([item])

    def insert(self, i, val):
        self.beginInsertRows(QModelIndex(), i, i)
        self._list.insert(i, val)
        self._other_data.insert(i, _store())
        self.endInsertRows()

    def remove(self, val):
        i = self._list.index(val)
        self.__delitem__(i)

    def pop(self, i):
        item = self._list[i]
        self.__delitem__(i)
        return item

    def indexOf(self, value):
        return self._list.index(value)

    def clear(self):
        del self[:]

    def __len__(self):
        return len(self._list)

    def __contains__(self, value):
        return value in self._list

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __add__(self, iterable):
        new_list = PyListModel(list(self._list),
                               # method parent is overloaded in Model
                               QObject.parent(self),
                               flags=self._flags,
                               list_item_role=self.list_item_role,
                               supportedDropActions=self.supportedDropActions())
        # pylint: disable=protected-access
        new_list._other_data = list(self._other_data)
        new_list.extend(iterable)
        return new_list

    def __iadd__(self, iterable):
        self.extend(iterable)
        return self

    def __delitem__(self, s):
        if isinstance(s, slice):
            start, stop, _ = _as_contiguous_range(s, len(self))
            if not len(self) or start == stop:
                return
            self.beginRemoveRows(QModelIndex(), start, stop - 1)
        else:
            s = operator.index(s)
            s = len(self) + s if s < 0 else s
            self.beginRemoveRows(QModelIndex(), s, s)
        del self._list[s]
        del self._other_data[s]
        self.endRemoveRows()

    def __setitem__(self, s, value):
        if isinstance(s, slice):
            start, stop, step = _as_contiguous_range(s, len(self))
            self.__delitem__(slice(start, stop, step))

            if not isinstance(value, list):
                value = list(value)
            if len(value) == 0:
                return
            self.beginInsertRows(QModelIndex(), start, start + len(value) - 1)
            self._list[start:start] = value
            self._other_data[start:start] = (_store() for _ in value)
            self.endInsertRows()
        else:
            s = operator.index(s)
            s = len(self) + s if s < 0 else s
            self._list[s] = value
            self._other_data[s] = _store()
            self.dataChanged.emit(self.index(s), self.index(s))

    def reverse(self):
        self._list.reverse()
        self._other_data.reverse()
        self.dataChanged.emit(self.index(0), self.index(len(self) - 1))

    def sort(self, *args, **kwargs):
        indices = _argsort(self._list, *args, **kwargs)
        lst = [self._list[i] for i in indices]
        other = [self._other_data[i] for i in indices]
        for i, (new_l, new_o) in enumerate(zip(lst, other)):
            self._list[i] = new_l
            self._other_data[i] = new_o
        self.dataChanged.emit(self.index(0), self.index(len(self) - 1))

    def __repr__(self):
        return "PyListModel(%s)" % repr(self._list)

    def __bool__(self):
        return len(self) != 0

    def emitDataChanged(self, indexList):
        if isinstance(indexList, int):
            indexList = [indexList]

        #TODO: group indexes into ranges
        for ind in indexList:
            self.dataChanged.emit(self.index(ind), self.index(ind))

    ###########
    # Drag/drop
    ###########

    def supportedDropActions(self):
        return self._supportedDropActions

    def mimeTypes(self):
        return [self.MIME_TYPE]

    def mimeData(self, indexlist):
        if len(indexlist) <= 0:
            return None

        def itemData(row):
            # type: (int) -> Dict[int, Any]
            if row < len(self._other_data):
                return {key: val for key, val in self._other_data[row].items()
                        if isinstance(key, int)}
            else:
                return {}  # pragma: no cover

        items = [self[i.row()] for i in indexlist]
        itemdata = [itemData(i.row()) for i in indexlist]
        mime = QMimeData()
        mime.setData(self.MIME_TYPE, b'see properties: _items, _itemdata')
        mime.setProperty('_items', items)
        mime.setProperty('_itemdata', itemdata)
        return mime

    def dropMimeData(self, mime, action, row, column, parent):
        if action == Qt.IgnoreAction:
            return True  # pragma: no cover

        if not mime.hasFormat(self.MIME_TYPE):
            return False  # pragma: no cover

        items = mime.property('_items')
        itemdata = mime.property('_itemdata')

        if not items:
            return False  # pragma: no cover

        if row == -1:
            row = len(self)  # pragma: no cover

        self[row:row] = items
        for i, data in enumerate(itemdata):
            self.setItemData(self.index(row + i), data)
        return True


class ListSingleSelectionModel(QItemSelectionModel):
    """ Item selection model for list item models with single selection.

    Defines signal:
        - selectedIndexChanged(QModelIndex)

    """
    selectedIndexChanged = Signal(QModelIndex)

    def __init__(self, model, parent=None):
        super().__init__(model, parent)
        self.selectionChanged.connect(self.onSelectionChanged)

    def onSelectionChanged(self, new, _):
        index = list(new.indexes())
        if index:
            index = index.pop()
        else:
            index = QModelIndex()

        self.selectedIndexChanged.emit(index)

    def selectedRow(self):
        """ Return QModelIndex of the selected row or invalid if no selection.
        """
        rows = self.selectedRows()
        if rows:
            return rows[0]
        else:
            return QModelIndex()

    def select(self, index, flags=QItemSelectionModel.ClearAndSelect):
        if isinstance(index, int):
            index = self.model().index(index)
        return super().select(self, index, flags)


def select_row(view, row):
    """
    Select a `row` in an item view.
    """
    selmodel = view.selectionModel()
    selmodel.select(view.model().index(row, 0),
                    QItemSelectionModel.ClearAndSelect |
                    QItemSelectionModel.Rows)


class ModelActionsWidget(QWidget):
    def __init__(self, actions=None, parent=None,
                 direction=QBoxLayout.LeftToRight):
        super().__init__(parent)
        self.actions = []
        self.buttons = []
        layout = QBoxLayout(direction)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        if actions is not None:
            for action in actions:
                self.addAction(action)
        self.setLayout(layout)

    def actionButton(self, action):
        if isinstance(action, QAction):
            button = QToolButton(self)
            button.setDefaultAction(action)
            return button
        elif isinstance(action, QAbstractButton):
            return action

    def insertAction(self, ind, action, *args):
        button = self.actionButton(action)
        self.layout().insertWidget(ind, button, *args)
        self.buttons.insert(ind, button)
        self.actions.insert(ind, action)
        return button

    def addAction(self, action, *args):
        return self.insertAction(-1, action, *args)



from unittest import mock
from typing import Sequence

from AnyQt.QtCore import Qt, QUrl, QPoint, QMimeData, QPointF
from AnyQt.QtGui import QDropEvent, QDragEnterEvent
from AnyQt.QtWidgets import QApplication, QWidget

from orangecanvas.registry import WidgetRegistry, WidgetDescription
from orangecanvas.document.interactions import PluginDropHandler, EntryPoint
from orangecanvas.document.schemeedit import SchemeEditWidget

from orangewidget.tests.base import GuiTest
from orangewidget.settings import Setting
from orangewidget.widget import OWBaseWidget
from orangewidget.workflow.widgetsscheme import WidgetsScheme
from orangewidget.workflow.drophandler import (
    SingleFileDropHandler, SingleUrlDropHandler, UrlsDropHandler,
    FilesDropHandler
)


class Widget(OWBaseWidget):
    name = "eman"
    param = Setting("")


class WidgetSingleUrlDropHandler(SingleUrlDropHandler):
    WIDGET = Widget

    def canDropUrl(self, url: QUrl) -> bool:
        return True

    def parametersFromUrl(self, url: QUrl) -> 'Dict[str, Any]':
        return {"param": url.toString()}


class WidgetSingleFileDropHandler(SingleFileDropHandler):
    WIDGET = Widget

    def canDropFile(self, path: str) -> bool:
        return True

    def parametersFromFile(self, path: str) -> 'Dict[str, Any]':
        return {"param": path}


class WidgetUrlsDropHandler(UrlsDropHandler):
    WIDGET = Widget

    def canDropUrls(self, urls: Sequence[QUrl]) -> bool:
        return True

    def parametersFromUrls(self, urls: Sequence[QUrl]) -> 'Dict[str, Any]':
        return {"param": [url.toString() for url in urls]}


class WidgetFilesDropHandler(FilesDropHandler):
    WIDGET = Widget

    def canDropFiles(self, files: Sequence[str]) -> bool:
        return True

    def parametersFromFiles(self, paths: Sequence[str]) -> 'Dict[str, Any]':
        return {"param": list(paths)}


def mock_iter_entry_points(module, attr):
    return mock.patch.object(
        PluginDropHandler, "iterEntryPoints",
        lambda _: [EntryPoint("AA", f"{module}:{attr}", "foo")]
    )


class TestDropHandlers(GuiTest):
    def setUp(self):
        super().setUp()
        reg = WidgetRegistry()
        reg.register_widget(
            WidgetDescription(**Widget.get_widget_description())
        )
        self.w = SchemeEditWidget()
        self.w.setRegistry(reg)
        self.w.resize(300, 300)
        self.w.setScheme(WidgetsScheme())
        self.w.setDropHandlers([PluginDropHandler()])

    def tearDown(self):
        self.w.scheme().clear()
        del self.w
        super().tearDown()

    @mock_iter_entry_points(__name__, WidgetSingleUrlDropHandler.__name__)
    def test_single_file_drop(self):
        w = self.w
        workflow = w.scheme()
        view = w.view()
        mime = QMimeData()
        mime.setUrls([QUrl("file:///foo/bar")])
        dragDrop(view.viewport(), mime, QPoint(10, 10))
        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].properties,
                         {"param": "file:///foo/bar"})

    @mock_iter_entry_points(__name__, WidgetSingleFileDropHandler.__name__)
    def test_single_url_drop(self):
        w = self.w
        workflow = w.scheme()
        view = w.view()
        mime = QMimeData()
        mime.setUrls([QUrl("file:///foo/bar")])
        dragDrop(view.viewport(), mime, QPoint(10, 10))
        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].properties,
                         {"param": "/foo/bar"})

    @mock_iter_entry_points(__name__, WidgetUrlsDropHandler.__name__)
    def test_urls_drop(self):
        w = self.w
        workflow = w.scheme()
        mime = QMimeData()
        mime.setUrls([QUrl("file:///foo/bar")] * 2)
        dragDrop(w.view().viewport(), mime, QPoint(10, 10))
        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].properties,
                         {"param": ["file:///foo/bar"] * 2})

    @mock_iter_entry_points(__name__, WidgetFilesDropHandler.__name__)
    def test_files_drop(self):
        w = self.w
        workflow = w.scheme()
        mime = QMimeData()
        mime.setUrls([QUrl("file:///foo/bar")] * 2)
        dragDrop(w.view().viewport(), mime, QPoint(10, 10))
        self.assertEqual(len(workflow.nodes), 1)
        self.assertEqual(workflow.nodes[0].properties,
                         {"param": ["/foo/bar"] * 2})


def dragDrop(
        widget: 'QWidget', mime: QMimeData, pos: QPoint = QPoint(-1, -1),
        action=Qt.CopyAction, buttons=Qt.LeftButton, modifiers=Qt.NoModifier
) -> bool:
    if pos == QPoint(-1, -1):
        pos = widget.rect().center()

    ev = QDragEnterEvent(pos, action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)
    if not ev.isAccepted():
        return False
    ev = QDropEvent(QPointF(pos), action, mime, buttons, modifiers)
    ev.setAccepted(False)
    QApplication.sendEvent(widget, ev)
    return ev.isAccepted()
